#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "pairings.h"
#include "funcs.h"
#include "commons.h"
#include "stats.h"
#include "arithm.h"

//----------------------------------------------------------------------------
typedef struct {
    //
    // PARI code copied over from src/basemath/FpX.c
    //
    GEN pol, p;
} FpX_muldata;
//----------------------------------------------------------------------------
static GEN
_FpXQ_sqr(void *data, GEN x) {
    //
    // PARI code copied over from src/basemath/FpX.c.
    // The code duplication was needed to make it possible to count operations.
    //
    FpX_muldata *D = (FpX_muldata*)data;
    return FPXQ_SQR(x, D->pol, D->p);
}
//----------------------------------------------------------------------------
static GEN
_FpXQ_mul(void *data, GEN x, GEN y) {
    //
    // PARI code copied over from src/basemath/FpX.c.
    // The code duplication was needed to make it possible to count operations.
    //
    FpX_muldata *D = (FpX_muldata*)data;
    return FPXQ_MUL(x,y, D->pol, D->p);
}
//----------------------------------------------------------------------------
static GEN
_ZMrow_ZC_mul_i(GEN x, GEN y, long lx, long i)
{
    //
    // PARI code copied over from src/basemath/FpV.c and slighlty reformated
    // The code duplication was needed to make it possible to count operations.
    //
    GEN c    = MULII(gcoeff(x, i, 1), gel(y, 1));
    GEN ZERO = gen_0;
    
    for (long k = 2; k < lx; k++) {
        GEN t = MULII(gcoeff(x, i, k), gel(y, k));
        if (t != ZERO) {
            c = ADDII(c, t);
        }
    }
    return c;
}
//----------------------------------------------------------------------------
static GEN
_FpM_FpC_mul_i(GEN x, GEN y, long lx, long l, GEN p)
{
    //
    // PARI code copied over from src/basemath/FpV.c and slighlty reformated
    // The code duplication was needed to make it possible to count operations.
    //
    GEN z = cgetg(l, t_COL);
    for (long i = 1; i < l; i++) {
      pari_sp av = avma;
      GEN c = _ZMrow_ZC_mul_i(x, y, lx, i);
      gel(z, i) = gerepileuptoint(av, modii(c, p));
    }
    return z;
}
//----------------------------------------------------------------------------
static int
_number_nonzero_coefs(GEN x)
{
    if (gequal0(x)) {
        return 0;
    }
    if (typ(x) == t_POL) {
        int nnz = 0;
        int lgx = lg(x);
        for (int i = 2; i < lgx; i++) {
            if (!gequal0(gel(x, i))) {
                nnz++;
            }
        }
        return nnz;
    }
    return 0;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_pow(GEN x, GEN n, GEN pol, GEN p) {
    //
    // PARI code copied over from src/basemath/FpX.c and slighlty simplified
    // (memory cleaning is _known_ to occur elsewhere).
    // The code duplication was needed to make it possible to count operations.
    //
    FpX_muldata D;
    long s = signe(n);
    GEN y;

    if (!s) {
        return pol_1(varn(x));
    }
    if (is_pm1(n)) {
        return (s < 0)? 
        FPXQ_INV(x,pol,p): ZX_copy(x);
    }
    if (!is_bigint(p)) {
        ulong pp = p[2];
        pol = ZX_to_Flx(pol, pp);
        x   = ZX_to_Flx(x,   pp);
        y   = Flx_to_ZX(Flxq_pow(x, n, pol, pp) );
    
    } else {
        D.pol = pol;
        D.p   = p;
        if (s < 0) {
            x = FPXQ_INV(x,pol,p);
        }
        y = gen_pow(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
    }
    return y;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_pow_simple(GEN base, GEN power, GEN T, GEN p) {
    //
    // Returns base^power in F_{p^k}.
    // T is the irreducible polynomial (in Fp) used to constructF_{q^k}.
    //
    // base and T are ZXs.
    // power and p are t_INTs.
    //
    GEN res = base;
    long i = expi(power) - 1;

    while (i >= 0) {
        res = FPXQ_SQR(res, T, p);
        if (_get_bit(power, i)) {
            res = FPXQ_MUL(res, base, T, p);
        }
        i--;
    }
    return res;
}
//----------------------------------------------------------------------------
GEN _shallow_ZX_to_FpC(GEN fpx, int d)
{
    //
    // Converts and returns (as a shallow copy) the ZX polynomial fpx as
    // a t_COL of dimension d, whose (i+1)-th component is the term of degree
    // i. If d is less than (degpol(fpx) - 1) the higher degree components
    // are just ignored.
    //
    GEN res = cgetg(d + 1, t_COL);
    int deg = degpol(fpx);
    if ((deg + 1) > d) {
        deg = d - 1;
    }
    for (int i = 0; i <= deg; i++) {
        res[i + 1] = fpx[i + 2];
    }
    for (int i = deg + 1; i < d; i++) {
        gel(res, i + 1) = stoi(0);
    }
    return res;
}
//----------------------------------------------------------------------------
GEN _shallow_FpC_to_ZX(GEN fpc, int vn)
{   
    //
    // Converts and returns (as a shallow copy) the t_COL fpc as a ZX
    // polynomial of degree at most (lg(fpc) - 2) in the variable number vn.
    //
    int rd = 0;
    int d = lg(fpc) - 1;
    for (int i = d; i > 0; i--) {
        if (!equalis((GEN)fpc[i], 0)) {
            rd = i;
            break;
        }
    }
    if (rd == 0) {
        GEN res = mkpoln(1, stoi(0));
        setvarn(res, vn);
        return res;
    }
    GEN res = cgetg(rd + 2, t_POL);
    for (int i = rd - 1; i >= 0; i--) {
        res[i + 2] = fpc[i + 1];
    }
    res[1] = evalvarn(0);
    setsigne(res, 1);
    setvarn(res, vn);
    return res;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_frobenius_map(GEN Q, GEN p)
{    
    int degQ = degpol(Q);

    GEN map = cgetg(degQ + 1, t_MAT);
    for (int i = 1; i < degQ + 1; i++) {
        gel(map, i) = cgetg(degQ + 1, t_COL);
    }
    GEN x    = RgX_to_FpX(mkpoln(2, gen_1, gen_0), p);
    GEN xpow = Rg_to_FpXQ(gen_1, Q, p);
    GEN fpow = FPXQ_POW(xpow, p, Q, p);
    
    for (int i = 1; i < degQ + 1; i++) {
        GEN col = gel(map, i);
        int dentry = degpol(fpow);
        for (int j = 0; j < dentry + 1; j++) {
            gel(col, j + 1) = gcopy((GEN)fpow[2 + j]);
        }
        for (int j = dentry + 1; j < degQ; j++) {
            gel(col, j + 1) = stoi(0);
        }
        xpow = FPXQ_MUL(xpow, x, Q, p);
        fpow = FPXQ_POW(xpow, p, Q, p);
    }
    map = RgM_to_FpM(map, p);
    return map;
}
//----------------------------------------------------------------------------
GEN
_FpM_mul(GEN x, GEN y, GEN p)
{
    //
    // PARI code copied over from src/basemath/FpV.c and slighlty reformated
    // (PARI's original developpers had to pay money for each space and newline
    // they used).
    // The code duplication was needed to make it possible to count operations.
    //
    long lx = lg(x);
    long ly = lg(y);

    if (ly == 1) {
        return cgetg(1, t_MAT);
    }
    if (lx == 1) {
        return zeromat(0, ly - 1);
    }
    long l = lg(x[1]);
    GEN  z = cgetg(ly, t_MAT);
    
    for (long j = 1; j < ly; j++) {
        gel(z, j) = _FpM_FpC_mul_i(x, gel(y, j), lx, l, p);
    }
    return z;
}
//----------------------------------------------------------------------------
GEN
_FpM_FpC_mul(GEN x, GEN y, GEN p)
{
  long lx = lg(x);
  if (lx == 1) {
      return cgetg(1,t_COL);
  } else {
      return _FpM_FpC_mul_i(x, y, lx, lg(x[1]), p);
  }
}
//----------------------------------------------------------------------------
GEN
_FpM_FpXQ_mul(GEN M, GEN a, GEN Q, GEN p)
{
    int degQ = degpol(Q);
    int vn   = varn(a);
    
    GEN acol = _shallow_ZX_to_FpC(a, degQ);    
    GEN fcol = _FpM_FpC_mul(M, acol, p);
    GEN f    = _shallow_FpC_to_ZX(fcol, vn);
    
    return f;
}
//----------------------------------------------------------------------------
int
_FpXQ_issquare(GEN x, GEN T, GEN p)
{
    //
    // PARI code copied over from src/basemath/FpV.c and slighlty reformated
    // The code duplication was needed to make it possible to count operations.
    //
    if (lg(x) == 2 || equalui(2, p)) {
        return 1;
    }
    pari_sp av = avma;
    GEN  m   = diviiexact(subis(powiu(p, degpol(T)), 1), subis(p,1));
    GEN  z   = constant_term(FPXQ_POW(x, m, T, p));
    long res = (kronecker(z, p) == 1);
    avma = av;
    return res;
}
//----------------------------------------------------------------------------
GEN
_matid_FpM(unsigned long n)
{
    GEN M = cgetg(n + 1,t_MAT);
    for (unsigned long i = 1; i < n + 1; i++) {
        gel(M, i) = cgetg(n + 1, t_COL);
        GEN col = gel(M, i);
        for (unsigned long i = 1; i < n + 1; i++) {
            gel(col, i) = stoi(0);
        }
    }
    for (unsigned long i = 1; i < n + 1; i++) {
        GEN col = gel(M, i);
        gel(col, i) = stoi(1);
    }
    return M;
}
//----------------------------------------------------------------------------
int _nfrob_powers_to_compute(pairing_data_t* const pdata) {
    int nf = (pdata->embed_deg / 2);
    switch (pdata->family) {
        case MNT:
            nf = 3;
            break;

        case BN:
            nf = 6;     
            break;
        
        case FREEMAN:
            nf = 5;            
            break;

        case OPTI_7_1333:
            nf = 5;
            break;
        
        case OPTI_11_1200:
            nf = 9;            
            break;

        case DUAN_ET_AL_15_1500:
            nf = 7;            
            break;

        case OPTI_15_1500:
            nf = 7;       
            break;
    
        case OPTI_16_1250:
            nf = 8;  
            break;
        
        case OPTI_17_1188:
            nf = 16;
            break;
        
        case OPTI_17_1125:
            nf = 16;
            break;
        
        case OPTI_18_1333:
            nf = 9;            
            break;

        case OPTI_19_1111:
            nf = 17;
            break;
                
        case OPTI_24_1250:
            nf = 12;
            break;

        case OPTI_25_1300:
            nf = 19;        
            break;

        case OPTI_26_1166:
            nf = 13;            
            break;
        
        case OPTI_27_1111:
            nf = 17;          
            break;
        
        default:
            break;
    }
    if (nf < MIN_NFROB_POWERS) {
        nf = MIN_NFROB_POWERS;
    }
    if (nf > MAX_NFROB_POWERS) {
        nf = MAX_NFROB_POWERS;
    }
    return nf;
}
//----------------------------------------------------------------------------
int
_FpXQ_compute_frobenius_powers(pairing_data_t* const pdata)
{
    IN();
    //
    // This function computes the matrices corresponding to the Frobenius
    // powers p^i with i = 0 .. pdata->embed_deg/2. Note that this is more
    // than what we strictly need for a given pairing but these are
    // precomputations and they are relatively fast anyway. Plus, this makes
    // for a much more simple (and thus maintainable) code.
    //
    int k = pdata->embed_deg;
    GEN p = pdata->charac;
    GEN t = pdata->irredpol;
    
    if (pdata->frobenius_powers[1] != NULL) {
        return SUCCESS;
    }
    pdata->frobenius_powers[0] = gclone(_matid_FpM(k));
    pdata->frobenius_powers[1] = gclone(_FpXQ_frobenius_map(t, p));

    int nf = _nfrob_powers_to_compute(pdata);
    GEN frob = pdata->frobenius_powers[1];
    GEN fpow = pdata->frobenius_powers[1];
    for (int i = 2; i <= nf; i++) {
        fpow = FPM_MUL(fpow, frob, p);
        pdata->frobenius_powers[i] = gclone(fpow);
    }
    OUT();
    return SUCCESS;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_point_frobenius_power(pairing_data_t* const pdata, GEN P, int i) {
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
    if (i > _nfrob_powers_to_compute(pdata)) {
        ERR("This Frobenius power was too large to be pre-computed");
    }
    if (ell_is_inf(P)) {
        return gcopy(P);
    }
    if (typ(gel(P, 1)) == t_INT) {
        //
        // The point P lives on E(F_p) -- simply return the same point.
        //
        return gcopy(P);
    }
    //
    // From now on, it is assumed that the point P lives on E(Fp^k).
    //
    GEN  t    = pdata->irredpol;
    GEN  p    = pdata->charac;
    GEN  frob = pdata->frobenius_powers[i];
    
    GEN PIP = gcopy(P);
    for (int j = 1; j < lg(PIP); j++) {
        gel(PIP, j) = FPM_FPXQ_MUL(frob, gel(PIP, j), t, p);
    }
    return PIP;    
}
//----------------------------------------------------------------------------
GEN
_FpXQ_frobenius_power(pairing_data_t* const pdata, GEN X, int i) {
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
    if (i > _nfrob_powers_to_compute(pdata)) {
        ERR("This Frobenius power was too large to be pre-computed");
    }
    if (typ(X) == t_INT) {
        //
        // X is in F_p -- simply return a copy.
        //
        return gcopy(X);
    }
    //
    // From now on, it is assumed that the X is in F_{p^k}.
    //
    GEN  t    = pdata->irredpol;
    GEN  p    = pdata->charac;
    GEN  frob = pdata->frobenius_powers[i];
    
    GEN PIP = FPM_FPXQ_MUL(frob, X, t, p);
    return PIP;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_red_2_safe(GEN a, GEN t, GEN p)
{    
    //
    // t must be of the form x^k + b 
    // a must be of the form c_{2k}.x^{2k} + c_{2k-1}.x^{2k-1} + ... + c_{0} 
    //
    int dega = degpol(a);
    int degt = degpol(t);
    int vn   = varn(a);
    
    if (dega < degt) {
        return gcopy(a);
    }
    GEN b    = gel(t, 2);    
    GEN res  = gcopy(a);
    GEN resc = res + 2;
    
    int degres = degpol(res);
    while (degres > 2 * (degt - 1)) {
        int stop = degres - degt;
        for (int i = degres; i > stop; i--) {
            GEN w  = Fp_mul(b,  gel(resc, i), p);
            GEN to = gel(resc, i - degt);
            gel(resc, i - degt) = Fp_sub(to, w, p);
        }
        degres -= degt;
    }
    
    for (int i = degres; i >= degt; i--) {
        GEN w  = Fp_mul(b, gel(resc, i), p);
        GEN to = gel(resc, i - degt);
        gel(resc, i - degt) = Fp_sub(to, w, p);
    }
    res[1] = evalvarn(0);
    setvarn(res, vn);

    res = normalizepol_lg(res, degt + 2);

    return res;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_red_2(GEN a, GEN t, GEN p)
{
    //
    // t must be of the form x^k + b 
    // a must be of the form c_{2k}.x^{2k} + c_{2k-1}.x^{2k-1} + ... + c_{0} 
    //
    int dega = degpol(a);
    int degt = degpol(t);
    int vn   = varn(a);
    
    if (dega < degt) {
        return gcopy(a);
    }
    if (dega > 2 * (degt - 1)) {
        ERR("dega > 2 * (degt - 1)");
        return gen_0;
    }
    GEN res  = cgetg(degt + 2, t_POL);
    GEN resc = res + 2;
    GEN b    = gel(t, 2);
    GEN ac   = a + 2;

    for (int i = 0; i < degt; i++) {
        gel(resc, i) = gcopy(gel(ac, i));
    }
        
    for (int i = dega; i >= degt; i--) {
        GEN w  = Fp_mul(b, gel(ac, i), p);
        GEN to = gel(ac, i - degt);
        gel(resc, i - degt) = Fp_sub(to, w, p);
    }
    res[1] = evalvarn(0);
    setvarn(res, vn);
    
    res = normalizepol_lg(res, degt + 2);

    return res;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_red_3_safe(GEN a, GEN t, GEN p)
{
    //
    // t must be of the form x^k + x + b 
    // a must be of the form c_{2k}.x^{2k} + c_{2k-1}.x^{2k-1} + ... + c_{0} 
    //
    int dega = degpol(a);
    int degt = degpol(t);
    int vn   = varn(a);
    
    if (dega < degt) {
        return gcopy(a);
    }
    GEN b    = gel(t, 2);    
    GEN res  = gcopy(a);
    GEN resc = res + 2;
    
    int degres = degpol(res);
    while (degres > 2 * (degt - 1)) {
        int stop = degres - degt;
        for (int i = degres; i > stop; i--) {
            GEN w  = Fp_mul(b, gel(resc, i), p);
            GEN to = gel(resc, i - degt);
            gel(resc, i - degt) = Fp_sub(to, w, p);
            gel(resc, i - degt + 1) = Fp_sub(gel(resc, i - degt + 1), gel(resc, i), p);
        }
        degres -= degt;
    }
    
    for (int i = degres; i >= degt; i--) {
        GEN w  = Fp_mul(b, gel(resc, i), p);
        GEN to = gel(resc, i - degt);
        gel(resc, i - degt) = Fp_sub(to, w, p);
        gel(resc, i - degt + 1) = Fp_sub(gel(resc, i - degt + 1), gel(resc, i), p);
    }
    res[1] = evalvarn(0);
    setvarn(res, vn);
    
    res = normalizepol_lg(res, degt + 2);
    
    return res;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_red_3(GEN a, GEN t, GEN p)
{
    //
    // t must be of the form x^k + x + b 
    // a must be of the form c_{2k}.x^{2k} + c_{2k-1}.x^{2k-1} + ... + c_{0} 
    //
    int dega = degpol(a);
    int degt = degpol(t);
    int vn   = varn(a);
    
    if (dega < degt) {
        return gcopy(a);
    }
    if (dega > 2 * (degt - 1)) {
        ERR("dega > 2 * (degt - 1)");
        return gen_0;
    }
    GEN res  = cgetg(degt + 2, t_POL);
    GEN resc = res + 2;
    GEN b    = gel(t, 2);
    GEN ac   = a + 2;

    for (int i = 0; i < degt; i++) {
        gel(resc, i) = gcopy(gel(ac, i));
    }
        
    for (int i = dega; i >= degt; i--) {
        GEN w  = Fp_mul(b, gel(ac, i), p);
        GEN to = gel(ac, i - degt);
        gel(resc, i - degt)     = Fp_sub(to, w, p);
        gel(resc, i - degt + 1) = Fp_sub(gel(resc, i - degt + 1), gel(ac, i), p);
    }
    res[1] = evalvarn(0);
    setvarn(res, vn);
    
    res = normalizepol_lg(res, degt + 2);

    return res;
}
//----------------------------------------------------------------------------
GEN
_FpXQ_red(GEN a, GEN t, GEN p)
{
    int nnz = _number_nonzero_coefs(t);
    switch (nnz) {
        case 2:
            return _FpXQ_red_2_safe(a, t, p);
            break;
        case 3:
            return _FpXQ_red_3_safe(a, t, p);
            break;
        default:
            return FpXQ_red(a, t, p);
            break;
    }
}
//----------------------------------------------------------------------------
