#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include "apip_config.h"
#include "pairings.h"
#include "funcs.h"
#include "commons.h"
#include "stats.h"
#include "arithm.h"
#include "affine.h"
#include "jacobian.h"
#include "proj.h"
#include "final_exp.h"

//----------------------------------------------------------------------------
//                      Pairing in characteristic > 3
//
// WARNING: All functions in this file assume the elliptic curve is given
//          by an equation of the form y^2 = x^3 + ax + b
//
// WARNING: Excepted when mentionned otherwise, all the functions in this
//          file are NOT memory safe!
//----------------------------------------------------------------------------
GEN final_exp_corr(pairing_data_t* pdata, GEN partial);
GEN opti_error(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_freeman(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_bn(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_5_1500(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_7_1333(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_8_1250(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_9_1333(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_11_1200(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_15_1500(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_16_1250(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_17_1188(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_17_1125(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_18_1333(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_19_1111(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_25_1300(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_27_1111(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_ate_26_1166(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_bn(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_9_1333(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_15_1500(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_16_1250(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_18_1333(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_24_1250(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_27_1111(pairing_data_t* pdata, GEN Q, GEN P);
GEN opti_twisted_duan_et_al_15_1500(pairing_data_t* pdata, GEN Q, GEN P);
//----------------------------------------------------------------------------
void
update_pairing_data_char_p(pairing_data_t* pdata) {
    if (pdata->f1_extdeg != 1) {
        //
        // To be implemented
        //
        return;
    }
    switch (pdata->coordsys) {
        case JACOBIAN:
            pariprintf("Using jacobian coordinates\n");
            switch (pdata->miller_method) {
                case STANDARD:
                    pariprintf("Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_jac_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_jac_fpxq_fp;
                    break;
                case NAF:
                    pariprintf("Using NAF Miller method\n");
                    pdata->miller_func_f1_f2 = naf_miller_jac_fp_fpxq;
                    pdata->miller_func_f2_f1 = naf_miller_jac_fpxq_fp;
                    break;
                case SLIDING_WINDOW:
                    //
                    // To be implemented
                    //
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_jac_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_jac_fpxq_fp;
                    break;
                case BOXALL_ET_AL:
                case BOXALL_ET_AL_NAF:
                    //
                    // To be implemented
                    //
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_aff_fpxq_fp;
                    break;
                default:
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_jac_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_jac_fpxq_fp;
            }
            pdata->point_mul_func_f1    = mul_jac_fp;
            pdata->point_add_func_f1    = add_jac_fp;
            pdata->point_double_func_f1 = double_jac_fp;

            pdata->point_mul_func_f2    = mul_jac_fpxq;
            pdata->point_add_func_f2    = add_jac_fpxq;
            pdata->point_double_func_f2 = double_jac_fpxq;

            pdata->eval_vert_f1_f2 = eval_vert_jac_fp_fpxq;
            pdata->eval_vert_f2_f1 = eval_vert_jac_fpxq_fp;

            pdata->update_chord_f1_f2   = update_chord_jac_fp_fpxq;
            pdata->update_chord_f2_f1   = update_chord_jac_fpxq_fp;
            pdata->update_tangent_f1_f2 = update_tangent_jac_fp_fpxq;
            pdata->update_tangent_f2_f1 = update_tangent_jac_fpxq_fp;
            break;

        case PROJECTIVE:
            pariprintf("Using projective coordinates\n");
            switch (pdata->miller_method) {
                case STANDARD:
                    pariprintf("Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_proj_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_proj_fpxq_fp;
                    break;
                case NAF:
                    pariprintf("Using NAF Miller method\n");
                    pdata->miller_func_f1_f2 = naf_miller_proj_fp_fpxq;
                    pdata->miller_func_f2_f1 = naf_miller_proj_fpxq_fp;
                    break;
                case SLIDING_WINDOW:
                    //
                    // To be implemented
                    //
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_proj_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_proj_fpxq_fp;
                    break;
                case BOXALL_ET_AL:
                case BOXALL_ET_AL_NAF:
                    //
                    // To be implemented
                    //
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_aff_fpxq_fp;
                    break;
                default:
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_proj_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_proj_fpxq_fp;
            }
            pdata->point_mul_func_f1    = mul_proj_fp;
            pdata->point_add_func_f1    = add_proj_fp;
            pdata->point_double_func_f1 = double_proj_fp;

            pdata->point_mul_func_f2    = mul_proj_fpxq;
            pdata->point_add_func_f2    = add_proj_fpxq;
            pdata->point_double_func_f2 = double_proj_fpxq;

            pdata->eval_vert_f1_f2 = eval_vert_proj_fp_fpxq;
            pdata->eval_vert_f2_f1 = eval_vert_proj_fpxq_fp;

            pdata->update_chord_f1_f2   = update_chord_proj_fp_fpxq;
            pdata->update_chord_f2_f1   = update_chord_proj_fpxq_fp;
            pdata->update_tangent_f1_f2 = update_tangent_proj_fp_fpxq;
            pdata->update_tangent_f2_f1 = update_tangent_proj_fpxq_fp;

            break;

        case AFFINE:
            pariprintf("Using affine coordinates\n");
            switch (pdata->miller_method) {
                case STANDARD:
                    pariprintf("Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_aff_fpxq_fp;
                    break;
                case NAF:
                    pariprintf("Using NAF Miller method\n");
                    pdata->miller_func_f1_f2 = naf_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = naf_miller_aff_fpxq_fp;
                    break;
                case SLIDING_WINDOW:
                    //
                    // To be implemented
                    //
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_aff_fpxq_fp;
                    break;
                case BOXALL_ET_AL:
                    pariprintf("Using Boxall et al's Miller variant\n");
                    pdata->miller_func_f1_f2 = boxalletal_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = boxalletal_miller_aff_fpxq_fp;
                    break;
                case BOXALL_ET_AL_NAF:
                    pariprintf("Using Boxall et al's Miller variant with NAF\n");
                    pdata->miller_func_f1_f2 = boxalletal_naf_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = boxalletal_naf_miller_aff_fpxq_fp;
                    break;
                default:
                    pariprintf("WARNING: Using standard Miller method\n");
                    pdata->miller_func_f1_f2 = std_miller_aff_fp_fpxq;
                    pdata->miller_func_f2_f1 = std_miller_aff_fpxq_fp;
            }
            pdata->point_mul_func_f1    = mul_aff_fp;
            pdata->point_add_func_f1    = add_aff_fp;
            pdata->point_double_func_f1 = double_aff_fp;

            pdata->point_mul_func_f2    = mul_aff_fpxq;
            pdata->point_add_func_f2    = add_aff_fpxq;
            pdata->point_double_func_f2 = double_aff_fpxq;

            pdata->eval_vert_f1_f2 = eval_vert_aff_fp_fpxq;
            pdata->eval_vert_f2_f1 = eval_vert_aff_fpxq_fp;

            pdata->update_chord_f1_f2   = update_chord_aff_fp_fpxq;
            pdata->update_chord_f2_f1   = update_chord_aff_fpxq_fp;
            pdata->update_tangent_f1_f2 = update_tangent_aff_fp_fpxq;
            pdata->update_tangent_f2_f1 = update_tangent_aff_fpxq_fp;

            break;
    }
    switch (pdata->family) {
        case BN:
            pdata->opti_ate_func     = opti_ate_bn;
            pdata->opti_twisted_func = opti_twisted_bn;
            break;

        case FREEMAN:
            pdata->opti_ate_func     = opti_ate_freeman;
            pdata->opti_twisted_func = opti_error;
            break;

        case DUAN_ET_AL_15_1500:
            pdata->opti_ate_func     = opti_error;
            pdata->opti_twisted_func = opti_twisted_duan_et_al_15_1500;
            break;

        case OPTI_5_1500:
            pdata->opti_ate_func     = opti_ate_5_1500;
            pdata->opti_twisted_func = opti_error;
            break;
        
        case OPTI_7_1333:
            pdata->opti_ate_func     = opti_ate_7_1333;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_9_1333:
            pdata->opti_ate_func     = opti_ate_9_1333;
            pdata->opti_twisted_func = opti_twisted_9_1333;
            break;

        case OPTI_11_1200:
            pdata->opti_ate_func     = opti_ate_11_1200;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_15_1500:
            pdata->opti_ate_func     = opti_ate_15_1500;
            pdata->opti_twisted_func = opti_twisted_15_1500;
            break;

        case OPTI_16_1250:
            pdata->opti_ate_func     = opti_ate_16_1250;
            pdata->opti_twisted_func = opti_twisted_16_1250;
            break;
        
        case OPTI_17_1188:
            pdata->opti_ate_func     = opti_ate_17_1188;
            pdata->opti_twisted_func = opti_error;
            break;
                
        case OPTI_17_1125:
            pdata->opti_ate_func     = opti_ate_17_1125;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_18_1333:
            pdata->opti_ate_func     = opti_ate_18_1333;
            pdata->opti_twisted_func = opti_twisted_18_1333;
            break;

        case OPTI_19_1111:
            pdata->opti_ate_func     = opti_ate_19_1111;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_24_1250:
            pdata->opti_ate_func     = opti_error;
            pdata->opti_twisted_func = opti_twisted_24_1250;
            break;

        case OPTI_25_1300:
            pdata->opti_ate_func     = opti_ate_25_1300;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_26_1166:
            pdata->opti_ate_func     = opti_ate_26_1166;
            pdata->opti_twisted_func = opti_error;
            break;

        case OPTI_27_1111:
            pdata->opti_ate_func     = opti_ate_27_1111;
            pdata->opti_twisted_func = opti_twisted_27_1111;
            break;

        default:
            pdata->opti_ate_func     = opti_error;
            pdata->opti_twisted_func = opti_error;
            break;
    }

    if (pdata->do_naive_exp) {
        pariprintf("Using naive exponentiation\n");
        pdata->final_exp_func = final_exp_naive_char_p;
    } else {
        pariprintf("Using optimized exponentiation\n");
        pdata->final_exp_func = final_exp_opti_char_p;
    }
    switch (pdata->twist_deg) {
        case 6:
        case 4:
        case 3:
        case 2:
            pariprintf("Using twist of degree %i\n", pdata->twist_deg);
            pdata->m = ugcd(pdata->embed_deg, pdata->twist_deg);
            pdata->e = pdata->embed_deg / pdata->m;
            break;
        default:
            pariprintf("No twist used\n");
            break;
    }
    pariflush();
}
//----------------------------------------------------------------------------
GEN
make_miller_inputs(const pairing_data_t* pdata, GEN P, GEN Q) {
    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int P_is_ffelt  = (typ(gel(P, 1)) == t_FFELT);
    int P_is_fpxq   = (gel(P, 1)[1]   == t_POL);
    int Q_is_intmod = (typ(gel(Q, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);

    if (P_is_intmod && Q_is_ffelt && Q_is_fpxq) {
        GEN PP = NULL;
        GEN QQ = NULL;

        GEN XPP = gel(gel(P, 1), 2);
        GEN YPP = gel(gel(P, 2), 2);
        GEN XQQ = FF_to_FpXQ(gel(Q,1));
        GEN YQQ = FF_to_FpXQ(gel(Q,2));

        switch (pdata->coordsys) {
            case AFFINE:
                PP = mkvec2(XPP, YPP);
                QQ = mkvec2(XQQ, YQQ);
                break;
            case PROJECTIVE:
            case JACOBIAN:
                PP = mkvecn(3, XPP, YPP, mkintn(1, 1));
                QQ = mkvecn(3, XQQ, YQQ, mkpoln(1, mkintn(1, 1)));
                break;
            default:
                ERR("coordinate system not specified");
                break;
        }
        return mkvec2(PP, QQ);
    }
    if (Q_is_intmod && P_is_ffelt && P_is_fpxq) {
        GEN PP = NULL;
        GEN QQ = NULL;

        GEN XPP = FF_to_FpXQ(gel(P,1));
        GEN YPP = FF_to_FpXQ(gel(P,2));
        GEN XQQ = gel(gel(Q, 1), 2);
        GEN YQQ = gel(gel(Q, 2), 2);

        switch (pdata->coordsys) {
            case AFFINE:
                PP = mkvec2(XPP, YPP);
                QQ = mkvec2(XQQ, YQQ);
                break;
            case PROJECTIVE:
            case JACOBIAN:
                PP = mkvecn(3, XPP, YPP, mkpoln(1, mkintn(1, 1)));
                QQ = mkvecn(3, XQQ, YQQ, mkintn(1, 1));
                break;
            default:
                ERR("coordinate system not specified");
                break;
        }
        return mkvec2(PP, QQ);
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
final_exp_corr(pairing_data_t* pdata, GEN partial)
{   
    GEN TT = pdata->irredpol;
    GEN  p = pdata->charac;
    
    if ((pdata->do_naive_exp == 0) 
            && (pdata->hardpow_method == SCOTT_ET_AL)
            && (pdata->cof != NULL)) {
        partial = FPXQ_POW(partial, pdata->cof, TT, p);        
    }
    return partial;
}
//----------------------------------------------------------------------------
GEN
tate_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;

    GEN   r = pdata->g1_order;
    GEN   p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN tate = pdata->miller_func_f1_f2(pdata, r, PP, QQ);

    PRINT_RATIOS();
    PRINT_STATS("Miller");
    RESET_STATS();

    if (pdata->do_reduce) {
        tate = pdata->final_exp_func(pdata, tate);
        tate = final_exp_corr(pdata, tate);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }
    GEN final = _FpXQ_to_FF(tate, TT, p);

    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
tate_pairing_im_ffelt_flxq(MAYBE_UNUSED pairing_data_t* pdata,
                           MAYBE_UNUSED GEN P, MAYBE_UNUSED GEN Q)
{
    // Not implemented yet...
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
tate_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{   IN();
    //
    // Returns the Tate pairing t(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);
    int Q_is_flxq   = (gel(Q, 1)[1]   == t_FF_Flxq);

    if (P_is_intmod && Q_is_ffelt) {
        if (Q_is_fpxq) {
            return tate_pairing_im_ffelt_fpxq(pdata, P, Q);
        }
        if (Q_is_flxq) {
            return tate_pairing_im_ffelt_flxq(pdata, P, Q);
        }
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
ate_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;

    GEN p = pdata->charac;
    GEN t = pdata->frob_trace;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN ate = pdata->miller_func_f2_f1(pdata, subis(t, 1), QQ, PP);

    PRINT_RATIOS();
    PRINT_STATS("Miller");
    RESET_STATS();

    if (pdata->do_reduce) {
        ate = pdata->final_exp_func(pdata, ate);
        ate = final_exp_corr(pdata, ate);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }

    GEN final = _FpXQ_to_FF(ate, TT, p);
    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
ate_pairing_im_ffelt_flxq(MAYBE_UNUSED pairing_data_t* pdata,
                          MAYBE_UNUSED GEN P, MAYBE_UNUSED GEN Q)
{
    // Not implemented yet...
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
ate_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    if (pdata->frob_trace == NULL) {
        ERR("Trace of the Frobenius is not specified");
        return NULL;
    }

    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);
    int Q_is_flxq   = (gel(Q, 1)[1]   == t_FF_Flxq);

    if (P_is_intmod && Q_is_ffelt) {
        if (Q_is_fpxq) {
            return ate_pairing_im_ffelt_fpxq(pdata, P, Q);
        }
        if (Q_is_flxq) {
            return ate_pairing_im_ffelt_flxq(pdata, P, Q);
        }
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
twisted_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();
    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;

    GEN   p = pdata->charac;
    GEN   t = pdata->frob_trace;
    ulong e = pdata->e;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN twisted = pdata->miller_func_f1_f2(pdata, powiu(subis(t, 1), e),PP,QQ);

    PRINT_RATIOS();
    PRINT_STATS("Miller");
    RESET_STATS();

    if (pdata->do_reduce) {
        twisted = pdata->final_exp_func(pdata, twisted);
        twisted = final_exp_corr(pdata, twisted);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }
    GEN final = _FpXQ_to_FF(twisted, TT, p);

    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
twisted_pairing_im_ffelt_flxq(MAYBE_UNUSED pairing_data_t* pdata,
                              MAYBE_UNUSED GEN P, MAYBE_UNUSED GEN Q)
{
    // Not implemented yet...
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
twisted_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Twisted ate pairing t(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    if (pdata->twist_deg == 0) {
        ERR("No twist degree specified");
        return NULL;
    }
    if (pdata->frob_trace == NULL) {
        ERR("Trace of the Frobenius is not specified");
        return NULL;
    }

    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);
    int Q_is_flxq   = (gel(Q, 1)[1]   == t_FF_Flxq);

    if (P_is_intmod && Q_is_ffelt) {
        if (Q_is_fpxq) {
            return twisted_pairing_im_ffelt_fpxq(pdata, P, Q);
        }
        if (Q_is_flxq) {
            return twisted_pairing_im_ffelt_flxq(pdata, P, Q);
        }
    }
    ERR("incompatible argument type");
    return gen_1;

}
//----------------------------------------------------------------------------
GEN
opti_ate_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;
    GEN   p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);
    RESET_STATS();

    GEN opti = pdata->opti_ate_func(pdata, QQ, PP);

    PRINT_RATIOS();
    PRINT_STATS("Millers and lines");
    RESET_STATS();

    if (pdata->do_reduce) {
        opti = pdata->final_exp_func(pdata, opti);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }
    GEN final = _FpXQ_to_FF(opti, TT, p);

    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
opti_ate_pairing_im_ffelt_flxq(MAYBE_UNUSED pairing_data_t* pdata,
                          MAYBE_UNUSED GEN P, MAYBE_UNUSED GEN Q)
{
    // Not implemented yet...
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
opti_ate_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Optimal Ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    if (pdata->frob_trace == NULL) {
        ERR("Trace of the Frobenius is not specified");
        return NULL;
    }
    if (pdata->u == NULL) {
        ERR("Family parameter undefined - use apip_set_family_param function");
        return NULL;
    }

    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);
    int Q_is_flxq   = (gel(Q, 1)[1]   == t_FF_Flxq);

    if (P_is_intmod && Q_is_ffelt) {
        if (Q_is_fpxq) {
            return opti_ate_pairing_im_ffelt_fpxq(pdata, P, Q);
        }
        if (Q_is_flxq) {
            return opti_ate_pairing_im_ffelt_flxq(pdata, P, Q);
        }
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;
    GEN  p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);
    RESET_STATS();

    GEN opti = pdata->opti_twisted_func(pdata, PP, QQ);

    PRINT_RATIOS();
    PRINT_STATS("Millers and lines");
    RESET_STATS();

    if (pdata->do_reduce) {
        opti = pdata->final_exp_func(pdata, opti);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }

    GEN final = _FpXQ_to_FF(opti, TT, p);
    
    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_pairing_im_ffelt_flxq(MAYBE_UNUSED pairing_data_t* pdata,
                                   MAYBE_UNUSED GEN P, MAYBE_UNUSED GEN Q)
{
    // Not implemented yet...
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Optimal twisted ate pairing a(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // If the characteristic of the involved field is p > 3 then:
    //
    //     - P must be a point on E(F_p) of order r whose coordinates are
    //       given by t_INTMODs.
    //     - Q must be a point on E(F_{p^k}) of order r whose coordinates
    //       are given by t_FFELTs of format t_FF_FpXQ or t_FF_Flxq.
    //
    if (pdata->frob_trace == NULL) {
        ERR("Trace of the Frobenius is not specified");
        return NULL;
    }
    if (pdata->u == NULL) {
        ERR("Family parameter undefined - use apip_set_family_param function");
        return NULL;
    }

    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);
    int Q_is_flxq   = (gel(Q, 1)[1]   == t_FF_Flxq);

    if (P_is_intmod && Q_is_ffelt) {
        if (Q_is_fpxq) {
            return opti_twisted_pairing_im_ffelt_fpxq(pdata, P, Q);
        }
        if (Q_is_flxq) {
            return opti_twisted_pairing_im_ffelt_flxq(pdata, P, Q);
        }
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
weil_pairing_ffelt_fpxq_im(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();
    //
    // Weil pairing does not involve denominator elimination and is only
    // performed in affine coordinates.
    //
    ulong          do_denom_elim = pdata->do_denom_elim;
    coord_system_t coordsys      = pdata->coordsys;

    pdata->do_denom_elim = 0;
    pdata->coordsys      = AFFINE;

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;

    GEN r = pdata->g1_order;
    GEN p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN light = pdata->miller_func_f1_f2(pdata, r, QQ, PP);

    PRINT_RATIOS();
    PRINT_STATS("Miller Light");
    RESET_STATS();

    GEN full = pdata->miller_func_f2_f1(pdata, r, PP, QQ);

    PRINT_STATS("Miller Full");
    RESET_STATS();

    //
    // weil(P, Q) = weil(Q, P)^(-1)
    //
    GEN weil = FpXQ_div(full, light, TT, p);
    if (mpodd(r)) {
        weil = FPX_NEG(weil, p);
    }
    GEN final = _FpXQ_to_FF(weil, TT, p);

    //
    // Restore pdata's state
    //
    pdata->do_denom_elim = do_denom_elim;
    pdata->coordsys      = coordsys;

    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
weil_pairing_im_ffelt_fpxq(pairing_data_t* pdata, GEN P, GEN Q)
{
    IN();
    //
    // Weil pairing does not involve denominator elimination and is only
    // performed in affine coordinates.
    //
    ulong          do_denom_elim = pdata->do_denom_elim;
    coord_system_t coordsys      = pdata->coordsys;

    pdata->do_denom_elim = 0;
    pdata->coordsys      = AFFINE;

    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;

    GEN r = pdata->g1_order;
    GEN p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN light = pdata->miller_func_f1_f2(pdata, r, PP, QQ);

    PRINT_RATIOS();
    PRINT_STATS("Miller Light");
    RESET_STATS();

    GEN full = pdata->miller_func_f2_f1(pdata, r, QQ, PP);

    PRINT_STATS("Miller Full");
    RESET_STATS();

    GEN weil  = FpXQ_div(light, full, TT, p);
    if (mpodd(r)) {
        weil = FPX_NEG(weil, p);
    }
    GEN final = _FpXQ_to_FF(weil, TT, p);

    //
    // Restore pdata's state
    //
    pdata->do_denom_elim = do_denom_elim;
    pdata->coordsys      = coordsys;

    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
weil_pairing_char_p(pairing_data_t* pdata, GEN P, GEN Q)
{
    //
    // Returns the Weil pairing w(P,Q) as a t_FFELT of F_{q^k}.
    //
    // pdata is a pointer to the pairing data.
    // Let's note:
    //     E = pdata->curve
    //     p = pdata->charac
    //     k = pdata->embed_deg
    //     r = pdata->g1_order
    // Then:
    //     P a point on E(F_p) of order r.
    //     Q a point on E(F_{p^k}) of order r.
    //
    // This function correctly cleans the stack upon completion.
    //
    // P and Q are two points of order r. Two cases are currently implemented:
    //
    // a) P is a point on E(F_p) of order r whose coordinates are given by
    //    t_INTMODs and Q is a point on E(F_{p^k}) of order r whose coordinates
    //    are given by t_FFELTs
    //
    // or
    //
    // b) P is a point on E(F_{p^k}) of order r whose coordinates are given
    //    by t_FFELTs and Q is a point on E(F_p) of order r whose coordinates
    //    are given by t_INTMODs
    //
    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int P_is_ffelt  = (typ(gel(P, 1)) == t_FFELT);
    int P_is_fpxq   = (gel(P, 1)[1]   == t_FF_FpXQ);
    int Q_is_intmod = (typ(gel(Q, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);

    if (P_is_intmod && Q_is_ffelt && Q_is_fpxq) {
        return weil_pairing_im_ffelt_fpxq(pdata, P, Q);
    }
    if (Q_is_intmod && P_is_ffelt && P_is_fpxq) {
        return weil_pairing_ffelt_fpxq_im(pdata, P, Q);
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
GEN
miller_im_ffelt_fpxq(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    IN();
    GEN inputs = make_miller_inputs(pdata, P, Q);

    GEN PP = gel(inputs, 1);
    GEN QQ = gel(inputs, 2);
    GEN TT = pdata->irredpol;
    GEN  p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN miller = pdata->miller_func_f1_f2(pdata, r, PP, QQ);

    PRINT_RATIOS();
    PRINT_STATS("Miller Light");
    RESET_STATS();

    if (pdata->do_reduce) {
        miller = pdata->final_exp_func(pdata, miller);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }
    GEN final = _FpXQ_to_FF(miller, TT, p);
    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
miller_ffelt_fpxq_im(pairing_data_t* pdata, GEN r, GEN Q, GEN P)
{
    IN();
    GEN inputs = make_miller_inputs(pdata, Q, P);

    GEN QQ = gel(inputs, 1);
    GEN PP = gel(inputs, 2);
    GEN TT = pdata->irredpol;
    GEN  p = pdata->charac;

    SET_STATS_EDEGREE(pdata->embed_deg);

    GEN miller = pdata->miller_func_f2_f1(pdata, r, QQ, PP);
    
    PRINT_RATIOS();
    PRINT_STATS("Miller Full");
    RESET_STATS();

    if (pdata->do_reduce) {
        miller = pdata->final_exp_func(pdata, miller);
        PRINT_STATS("Final Exp");
        RESET_STATS();
    }
    GEN final = _FpXQ_to_FF(miller, TT, p);
    OUT();
    return final;
}
//----------------------------------------------------------------------------
GEN
miller_char_p(pairing_data_t* pdata, GEN r, GEN P, GEN Q)
{
    int P_is_intmod = (typ(gel(P, 1)) == t_INTMOD);
    int P_is_ffelt  = (typ(gel(P, 1)) == t_FFELT);
    int P_is_fpxq   = (gel(P, 1)[1]   == t_FF_FpXQ);
    int Q_is_intmod = (typ(gel(Q, 1)) == t_INTMOD);
    int Q_is_ffelt  = (typ(gel(Q, 1)) == t_FFELT);
    int Q_is_fpxq   = (gel(Q, 1)[1]   == t_FF_FpXQ);

    if (P_is_intmod && Q_is_ffelt && Q_is_fpxq) {
        return miller_im_ffelt_fpxq(pdata, r, P, Q);
    }
    if (Q_is_intmod && P_is_ffelt && P_is_fpxq) {
        return miller_ffelt_fpxq_im(pdata, r, P, Q);
    }
    ERR("incompatible argument type");
    return gen_1;
}
//----------------------------------------------------------------------------
void
compute_frobenius_powers_char_p(pairing_data_t* pdata)
{
    if (pdata->frobenius_powers[1] == NULL) {
        (void)_FpXQ_compute_frobenius_powers(pdata);
    }
}
//----------------------------------------------------------------------------
GEN
frobenius_power_char_p(pairing_data_t* pdata, GEN X, unsigned int i)
{
    //
    // If X is an element of F_{p} given as a t_INTMOD, returns the i-th power
    // Frobenius of X as a t_INTMOD (basically a simple copy of X).
    //
    // If X is an element of F_{p^k} given as a t_FFELT, returns the i-th power
    // Frobenius of X as a t_FFELT.
    //
    int X_is_intmod = (typ(X) == t_INTMOD);
    int X_is_ffelt  = (typ(X) == t_FFELT);
    int X_is_fpxq   = (gel(X, 1) == t_FF_FpXQ);
    
    if (X_is_intmod) {
        return gcopy(X);
    }
    if (X_is_ffelt && X_is_fpxq) {
        GEN TT  = pdata->irredpol;
        GEN  p  = pdata->charac;
        GEN XX  = FF_to_FpXQ(X);
        GEN res = _FpXQ_frobenius_power(pdata, XX, i);
        return _FpXQ_to_FF(res, TT, p);;
    }
    ERR("incompatible argument type");
    return NULL;
}
//----------------------------------------------------------------------------
void
print_frobenius_matrix_char_p(pairing_data_t* const pdata, unsigned int i)
{
    if ((i > MAX_NFROB_POWERS) || (pdata->frobenius_powers[i] == NULL)) {
        pariprintf("%s: matrix not computed\n", __func__);
        pariflush();
    }
    pariprintf("%s\n", GENtostr(pdata->frobenius_powers[i]));
    pariflush();
}
//----------------------------------------------------------------------------
GEN
opti_error(MAYBE_UNUSED pairing_data_t* pdata, MAYBE_UNUSED GEN Q,
           MAYBE_UNUSED GEN P) {
    ERR("Optimal pairing not implemented for this curve family");
    return NULL;
}
//----------------------------------------------------------------------------
GEN
opti_ate_27_1111(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [X, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //     f = f_{X,Q} * l_{p^10*Q, X*Q} / v_{(X + p^10)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // Compute [p^10]*Q
    //
    GEN Q10 = _FpXQ_point_frobenius_power(pdata, Q, 10);

    //
    // Computing f_{X, Q}
    //
    GEN FXQ = pdata->miller_func_f2_f1(pdata, X, Q, P);

    //
    // Computing L = l_{p^10*Q, X*Q}(P) and V = v_{p^10*Q + X*Q}(P)
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q10, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_DIV(FPXQ_MUL(FXQ, L, TT, p), V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_26_1166(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [X^2,  X, 1,  0, 0,  0, 0,  0, 0,  0, 0, 0]
    //
    // yields the following optimal pairing function:
    //     f = f_{X^2,Q} * f_{X,Q}^p * l_{(p^2)*Q, X*p*Q}
    //                   * l_{(X*p + p^2)*Q, X^2*Q}
    //
    // to be evaluated at point P.
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // Computing f_{X, Q}^p and [X*p]*Q
    //
    GEN FX_Q   = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN FX_Q_P = _FpXQ_frobenius_power(pdata, FX_Q, 1);
    GEN X_Q    = gcopy(pdata->point);

    //
    // f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN Q_P2   = _FpXQ_point_frobenius_power(pdata, Q, 2);
    GEN XP_Q   = _FpXQ_point_frobenius_power(pdata, X_Q, 1);
    
    //
    // Computing l_{(p^2)*Q, X*p*Q} * l_{(X*p + p^2)*Q, X^2*Q}
    //
    pdata->point = gcopy(XP_Q);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q_P2, P);
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN opti = FPXQ_MUL(FX2_Q, FX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, pdata->num, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_25_1300(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // NOTE: This is actually the same function as opti_ate_19_1111
    //
    // Using the following short vector:
    //     [X^2, -X, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //    f = f_{X^2,Q} * f_{-X,Q}^p * l_{(p^2)*Q, -X*p*Q}
    //                  * l_{(p^2 - X*p)*Q, X^2*Q}
    //                  / (v_{(X^2 - X*p + p^2)*Q} * v_{(p^2 - X*p)*Q})
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // f_{-X,Q} is computed as 1 / (f{X, Q} * v{XQ})
    //
    GEN FX_Q  = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q   = gcopy(pdata->point);
    GEN FMX_Q = FPXQ_MUL(FX_Q, pdata->eval_vert_f2_f1(pdata, X_Q, P), TT, p);
    FMX_Q = FPXQ_INV(FMX_Q, TT, p);

    GEN MX_Q = gcopy(X_Q);
    gel(MX_Q, 2) = FPX_NEG(gel(X_Q, 2), p);
    
    GEN MX_Q_P  = _FpXQ_point_frobenius_power(pdata, MX_Q, 1);
    GEN FMX_Q_P = _FpXQ_frobenius_power(pdata, FMX_Q, 1);

    //
    // f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN Q_P2   = _FpXQ_point_frobenius_power(pdata,  Q, 2);

    //
    // Computing l_{(p^2)*Q, -X*p*Q} and v_{(p^2 - X*p)*Q}
    //
    pdata->point = gcopy(MX_Q_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q_P2, P);

    //
    // Computing l_{(p^2 - X*p)*Q, X^2*Q}  and v_{(X^2 - X*p + p^2)*Q
    //
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX2_Q, FMX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_18_1333(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [1, 3*x/7, 3*x/7 + 1, 0, -2*x/7, -2*x/7-1]
    //
    // yields the following optimal pairing function:
    //  f_{3*x/7,Q}^{p} * f_{3*x/7 + 1,Q}^{p^2}
    //  * f_{-2*x/7-1,Q}^{q^5} * f_{-2*x/7,Q}^{p^4}
    //  * l_{((3*x/7)*p + (3*x/7 + 1)*p^2 + (-2*x/7)*p^4 + (-2*x/7-1)*p^5)*Q, Q}
    //  * l_{((3*x/7 + 1)*p^2 + (-2*x/7)*p^4 + (-2*x/7-1)*p^5)*Q, (3*x/7)*p*Q}
    //  * l_{((-2*x/7)*p^4 + (-2*x/7-1)*p^5)*Q, (3*x/7 + 1) * p^2*Q}
    //  * l_{((-2*x/7-1) * p^5)*Q, (-2*x/7) * p^4*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X7 = DIVIUEXACT(X, 7);
    int vn = varn(gel(Q, 1));
    
    //
    // Computes f_{3*x/7,Q} as f_{x/7,Q}^{3} * f_{3, (x/7)*Q}
    //
    GEN FX7_Q    = pdata->miller_func_f2_f1(pdata, X7, Q, P);
    GEN X7_Q     = gcopy(pdata->point);
    GEN FX7_Q_2  = FPXQ_SQR(FX7_Q, TT, p);
    GEN FX7_Q_3  = FPXQ_MUL(FX7_Q, FX7_Q_2, TT, p);
    
    GEN F2_X7_Q  = pdata->miller_func_f2_f1(pdata, stoi(2), X7_Q, P);
    GEN _2X7_Q   = gcopy(pdata->point);
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, X7_Q, P);
    
    GEN _3X7_Q  = gcopy(pdata->point);
    GEN F3_X7_Q = FPXQ_MUL(pdata->num, F2_X7_Q, TT, p);
    GEN F3X7_Q  = FPXQ_MUL(F3_X7_Q, FX7_Q_3, TT, p);

    //
    // Computes f_{3*x/7+1,Q} as f_{3*x/7,Q} * l_{(3*x/7)*Q, Q}
    //
    pdata->point = gcopy(_3X7_Q);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN _3X7P_Q = gcopy(pdata->point);
    GEN F3X7P_Q = FPXQ_MUL(pdata->num, F3X7_Q, TT, p);
    
    //
    // Computes f_{2*x/7,Q} as f_{x/7,Q}^{2} * f_{2, (x/7)*Q}
    //
    GEN F2X7_Q  = FPXQ_MUL(F2_X7_Q, FX7_Q_2, TT, p);
    
    //
    // Computes f_{2*x/7+1,Q} as f_{2*x/7,Q} * l_{(2*x/7)*Q, Q}
    //
    pdata->point = gcopy(_2X7_Q);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN _2X7P_Q = gcopy(pdata->point);    
    GEN F2X7P_Q = FPXQ_MUL(pdata->num, F2X7_Q, TT, p);
    
    GEN F3X7_Q_P   = _FpXQ_frobenius_power(pdata, F3X7_Q,  1);
    GEN F3X7P_Q_P2 = _FpXQ_frobenius_power(pdata, F3X7P_Q, 2);
    GEN F2X7_Q_P4  = _FpXQ_frobenius_power(pdata, F2X7_Q,  4);
    GEN F2X7P_Q_P5 = _FpXQ_frobenius_power(pdata, F2X7P_Q, 5);
    
    GEN FINV = FPXQ_MUL(F2X7_Q_P4, F2X7P_Q_P5, TT, p);
    FINV = FPXQ_INV(FINV, TT, p);
    
    //
    // Computing line products
    //
    // Denote A = [3*x/7*p]*Q
    //        B = [(3*x/7 + 1)*q^2]*Q
    //        C = [-2*x/7*q^4]*Q
    //        D = [(-2*x/7-1)*q^5]*Q
    //
    GEN _M2X7_Q     = gcopy(_2X7_Q);
    gel(_M2X7_Q, 2) = FPX_NEG(gel(_2X7_Q, 2), p);
    
    GEN _M2X7M_Q     = gcopy(_2X7P_Q);
    gel(_M2X7M_Q, 2) = FPX_NEG(gel(_2X7P_Q, 2), p);
    
    GEN A = _FpXQ_point_frobenius_power(pdata,   _3X7_Q, 1);
    GEN B = _FpXQ_point_frobenius_power(pdata,  _3X7P_Q, 2);
    GEN C = _FpXQ_point_frobenius_power(pdata,  _M2X7_Q, 4);
    GEN D = _FpXQ_point_frobenius_power(pdata, _M2X7M_Q, 5);

    //
    // Compute the following line products
    //
    //    l_{D, C} * l_{C + D, B} * l_{B + C + D, A} * l_{A + B + C + D, Q}
    //
    pdata->point = gcopy(D);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, C, P);
    pdata->update_chord_f2_f1(pdata, B, P);
    pdata->update_chord_f2_f1(pdata, A, P);
    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN opti = pdata->num;

    opti = FPXQ_MUL(opti,   F3X7_Q_P, TT, p);
    opti = FPXQ_MUL(opti, F3X7P_Q_P2, TT, p);
    opti = FPXQ_MUL(opti,       FINV, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_19_1111(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // NOTE: This is actually the same function as opti_25_1111
    //
    // Using the following short vector:
    //     [X^2, -X, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //    f = f_{X^2,Q} * f_{-X,Q}^p * l_{(p^2)*Q, -X*p*Q}
    //                  * l_{(p^2 - X*p)*Q, X^2*Q}
    //                  / (v_{(X^2 - X*p + p^2)*Q} * v_{(p^2 - X*p)*Q})
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // f_{-X,Q} is computed as 1 / (f{X, Q} * v{XQ})
    //
    GEN FX_Q  = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q   = gcopy(pdata->point);
    GEN FMX_Q = FPXQ_MUL(FX_Q, pdata->eval_vert_f2_f1(pdata, X_Q, P), TT, p);
    FMX_Q = FPXQ_INV(FMX_Q, TT, p);

    GEN MX_Q = gcopy(X_Q);
    gel(MX_Q, 2) = FPX_NEG(gel(X_Q, 2), p);
    
    GEN MX_Q_P  = _FpXQ_point_frobenius_power(pdata, MX_Q, 1);
    GEN FMX_Q_P = _FpXQ_frobenius_power(pdata, FMX_Q, 1);

    //
    // f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN Q_P2   = _FpXQ_point_frobenius_power(pdata,  Q, 2);

    //
    // Computing l_{(p^2)*Q, -X*p*Q} and v_{(p^2 - X*p)*Q}
    //
    pdata->point = gcopy(MX_Q_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q_P2, P);

    //
    // Computing l_{(p^2 - X*p)*Q, X^2*Q}  and v_{(X^2 - X*p + p^2)*Q
    //
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX2_Q, FMX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_16_1250(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //    [A, B, C, D, E, F, G, H]
    //
    // yields the following optimal pairing function:
    //    f_{A,Q} * f_{B,Q}^{q} * f_{C,Q}^{q^2} * f_{D,Q}^{q^3}
    //      * f_{E,Q}^{q^4} * f_{F,Q}^{q^5} * f_{G,Q}^{q^6} * f_{H,Q}^{q^7}
    //      * l_{(B*q + C*q^2 + D*q^3 + E*q^4 + F*q^5 + G*q^6 + H*q^7)*Q, A*Q}
    //      * l_{(C*q^2 + D*q^3 + E*q^4 + F*q^5 + G*q^6 + H*q^7)*Q, B*q*Q}
    //      * l_{(D*q^3 + E*q^4 + F*q^5 + G*q^6 + H*q^7)*Q, C*q^2*Q}
    //      * l_{(E*q^4 + F*q^5 + G*q^6 + H*q^7)*Q, D*q^3*Q}
    //      * l_{(F*q^5 + G*q^6 + H*q^7)*Q, E*q^4*Q}
    //      * l_{(G*q^6 + H*q^7)*Q, F*q^5*Q}
    //      * l_{(H*q^7)*Q, G*q^6*Q}
    //
    // to be evaluated at point P.
    //
    // Values for x can only be 25 or 45 mod 70.
    //
    // When x = 25 mod 70, the following vector is used:
    //    [(2*x-15)/35, -(11*x-30)/35, -(2*x-1)/7, (x+10)/35, (2*x+5)/5,
    //     (8*x+10)/35, (2*x+6)/7, (17*x+30)/35]
    //
    // When x = 45 mod 70, the following vector is used:
    //    [(13*x-25)/35, (3*x+5)/35, -(17*x+5)/35, -(16*x+15)/35,
    //     -(4*x+30)/35, -(4*x+30)/35, (x-10)/35, -(x-5)/5]
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));
    GEN A, B, C, D, E, F, G, H;
    
    if (umodiu(X, 70) == 25) {
        GEN _2X  = ADDII(X, X);
        GEN _11X = MULUI(11, X);
        GEN _8X  = MULUI(8, X);
        GEN _17X = MULUI(17, X);
        A = DIVIUEXACT(ADDIS(_2X,  -15), 35);
        B = NEGI(DIVIUEXACT(ADDIS(_11X, -30), 35));
        C = NEGI(DIVIUEXACT(ADDIS(_2X,   -1), 7));
        D = DIVIUEXACT(ADDIS(X,     10), 35);
        E = DIVIUEXACT(ADDIS(_2X,    5), 5);
        F = DIVIUEXACT(ADDIS(_8X,   10), 35);
        G = DIVIUEXACT(ADDIS(_2X,    6), 7);
        H = DIVIUEXACT(ADDIS(_17X,  30), 35);
                
    } else {
        GEN _3X  = MULUI(3,      X);
        GEN _4X  = ADDII(_3X,    X);
        GEN _13X = MULUI(13,     X);
        GEN _16X = ADDII(_13X, _3X);
        GEN _17X = ADDII(_16X,   X);
        A = DIVIUEXACT(ADDIS(_13X, -25), 35);
        B = DIVIUEXACT(ADDIS(_3X,    5), 35);
        C = NEGI(DIVIUEXACT(ADDIS(_17X,  5), 35));
        D = NEGI(DIVIUEXACT(ADDIS(_16X, 15), 35));
        E = NEGI(DIVIUEXACT(ADDIS(_4X,  30), 35));
        F = E;
        G = DIVIUEXACT(ADDIS(X, -10), 35);
        H = NEGI(DIVIUEXACT(ADDIS(X, -5), 5));
    }
    GEN FA_Q = pdata->miller_func_f2_f1(pdata, A, Q, P);
    GEN A_Q  = gcopy(pdata->point);

    GEN FB_Q   = pdata->miller_func_f2_f1(pdata, B, Q, P);
    GEN FB_Q_P = _FpXQ_frobenius_power(pdata, FB_Q, 1);
    GEN B_Q    = gcopy(pdata->point);
    GEN B_Q_P  = _FpXQ_point_frobenius_power(pdata, B_Q, 1);

    GEN FC_Q    = pdata->miller_func_f2_f1(pdata, C, Q, P);
    GEN FC_Q_P2 = _FpXQ_frobenius_power(pdata, FC_Q, 2);
    GEN C_Q     = gcopy(pdata->point);
    GEN C_Q_P2  = _FpXQ_point_frobenius_power(pdata, C_Q, 2);

    GEN FD_Q    = pdata->miller_func_f2_f1(pdata, D, Q, P);
    GEN FD_Q_P3 = _FpXQ_frobenius_power(pdata, FD_Q, 3);
    GEN D_Q     = gcopy(pdata->point);
    GEN D_Q_P3  = _FpXQ_point_frobenius_power(pdata, D_Q, 3);

    GEN FE_Q    = pdata->miller_func_f2_f1(pdata, E, Q, P);
    GEN FE_Q_P4 = _FpXQ_frobenius_power(pdata, FE_Q, 4);
    GEN E_Q     = gcopy(pdata->point);
    GEN E_Q_P4  = _FpXQ_point_frobenius_power(pdata, E_Q, 4);
    
    GEN FF_Q    = pdata->miller_func_f2_f1(pdata, F, Q, P);
    GEN FF_Q_P5 = _FpXQ_frobenius_power(pdata, FF_Q, 5);
    GEN F_Q     = gcopy(pdata->point);
    GEN F_Q_P5  = _FpXQ_point_frobenius_power(pdata, F_Q, 5);

    GEN FG_Q    = pdata->miller_func_f2_f1(pdata, G, Q, P);
    GEN FG_Q_P6 = _FpXQ_frobenius_power(pdata, FG_Q, 6);
    GEN G_Q     = gcopy(pdata->point);
    GEN G_Q_P6  = _FpXQ_point_frobenius_power(pdata, G_Q, 6);

    GEN FH_Q    = pdata->miller_func_f2_f1(pdata, H, Q, P);
    GEN FH_Q_P7 = _FpXQ_frobenius_power(pdata, FH_Q, 7);
    GEN H_Q     = gcopy(pdata->point);
    GEN H_Q_P7  = _FpXQ_point_frobenius_power(pdata, H_Q, 7);
    
    pdata->point = gcopy(H_Q_P7);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f2_f1(pdata, G_Q_P6, P);
    pdata->update_chord_f2_f1(pdata, F_Q_P5, P);
    pdata->update_chord_f2_f1(pdata, E_Q_P4, P);
    pdata->update_chord_f2_f1(pdata, D_Q_P3, P);
    pdata->update_chord_f2_f1(pdata, C_Q_P2, P);
    pdata->update_chord_f2_f1(pdata, B_Q_P, P);
    pdata->update_chord_f2_f1(pdata, A_Q, P);

    GEN opti = pdata->num;
    
    opti = FPXQ_MUL(FA_Q,    opti, TT, p);
    opti = FPXQ_MUL(FB_Q_P,  opti, TT, p);
    opti = FPXQ_MUL(FC_Q_P2, opti, TT, p);
    opti = FPXQ_MUL(FD_Q_P3, opti, TT, p);
    opti = FPXQ_MUL(FE_Q_P4, opti, TT, p);
    opti = FPXQ_MUL(FF_Q_P5, opti, TT, p);
    opti = FPXQ_MUL(FG_Q_P6, opti, TT, p);
    opti = FPXQ_MUL(FH_Q_P7, opti, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_17_1188(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [X^2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //     f_{(X^2),Q} * l_{p*Q, X^2*Q} / v_{(X^2 + p)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    int vn = varn(gel(Q, 1));
    
    //
    // Computing f_{X^2,Q} and thus [X^2]Q
    //
    GEN FX2_Q  = pdata->miller_func_f2_f1(pdata, X2, Q, P);
    
    //
    // Computing l_{p*Q, X^2*Q} and v_{(X^2 + p)*Q}
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    GEN Q_P = _FpXQ_point_frobenius_power(pdata, Q, 1);

    pdata->update_chord_f2_f1(pdata, Q_P, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX2_Q, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_17_1125(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [X^3, 1]
    //
    // yields the following optimal pairing function:
    //     f_{X^3,Q} * l_{p*Q, X^3*Q} / v_{(X^3 + p)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X3 = MULII(X2, X);
    int vn = varn(gel(Q, 1));
    
    //
    // Computing f_{X^3,Q} and thus [X^3]Q
    //
    GEN FX3_Q = pdata->miller_func_f2_f1(pdata, X3, Q, P);

    //
    // Computing l_{p*Q, X^3*Q} and v_{(X^3 + p)*Q}
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    GEN Q_P = _FpXQ_point_frobenius_power(pdata, Q, 1);

    pdata->update_chord_f2_f1(pdata, Q_P, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX3_Q, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_15_1500(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [1, 0, 0, 0, X, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //     f_{X,Q}^{p^4} * l_{X*p^4*Q, Q} / v_{(1 + X*p^4)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // Computing f_{X, Q}^{p^4} and [X*p^4]*Q
    //
    GEN QP4   = _FpXQ_point_frobenius_power(pdata, Q, 4);
    GEN FXQP4 = pdata->miller_func_f2_f1(pdata, X, QP4, P);

    //
    // Computing l_{X*p^4*Q, Q} and v_{(1 + X*p^4)*Q}
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FXQP4, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_11_1200(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // NOTE: This is actually the same function as opti_25_1111
    //
    // Using the following short vector:
    //     [X^2, -X, 1, 0, 0, 0, 0, 0, 0, 0]
    //
    // yields the following optimal pairing function:
    //    f_{X^2,Q} * f_{-X,Q}^{q} * l_{(-X*q + q^2)*Q, X^2*Q}
    //    * l_{q^2*Q, -X*q*Q} / (v_{(X^2 + -X*q + q^2)*Q} * v_{(-X* q + q^2)*Q})
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // f_{-X,Q} is computed as 1 / (f{X, Q} * v{XQ})
    //
    GEN FX_Q  = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q   = gcopy(pdata->point);
    GEN FMX_Q = FPXQ_MUL(FX_Q, pdata->eval_vert_f2_f1(pdata, X_Q, P), TT, p);
    FMX_Q = FPXQ_INV(FMX_Q, TT, p);

    GEN MX_Q = gcopy(X_Q);
    gel(MX_Q, 2) = FPX_NEG(gel(X_Q, 2), p);
    
    GEN MX_Q_P  = _FpXQ_point_frobenius_power(pdata, MX_Q, 1);
    GEN FMX_Q_P = _FpXQ_frobenius_power(pdata, FMX_Q, 1);

    //
    // f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN Q_P2   = _FpXQ_point_frobenius_power(pdata,  Q, 2);

    //
    // Computing l_{(p^2)*Q, -X*p*Q} and v_{(p^2 - X*p)*Q}
    //
    pdata->point = gcopy(MX_Q_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q_P2, P);

    //
    // Computing l_{(p^2 - X*p)*Q, X^2*Q}  and v_{(X^2 - X*p + p^2)*Q
    //
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX2_Q, FMX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_9_1333(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [(X-2)/3, (X+1)/3, (X+1)/3, -(X+1)/3, -(X-2)/3, -(X-2)/3]
    //
    // yields the following optimal pairing function:
    //     f_{A,Q} * f_{B,Q}^{q} * f_{B,Q}^{q^2} * f_{-B,Q}^{q^3} * f_{-A,Q}^{q^4}
    //             * f_{-A,Q}^{q^5}
    //             * l_{[B*q + B*q^2 - B*q^3 - A*q^4 - A*q^5] Q, [A] Q}
    //             * l_{[B*q^2 - B*q^3 - A*q^4 - A*q^5] Q, [B*q] Q}
    //             * l_{[-B*q^3 - A*q^4 -A*q^5] Q, [B * q^2] Q}
    //             * l_{[-A*q^4 - A*q^5] Q, [-B*q^3] Q}
    //             * l_{[-A*q^5] Q, [-A*q^4] Q}
    //             / (  v_{[A + B*q + B*q^2 - B*q^3 - A*q^4 - A*q^5] Q} 
    //                * v_{[B* q + B*q^2 - B*q^3 - A*q^4 - A*q^5] Q}
    //                * v_{[B*q^2 - B*q^3 - A*q^4 - A*q^5] Q}
    //                * v_{[-B*q^3 - A*q^4 - A*q^5] Q}
    //                * v_{[-A*q^4 - A*q^5] Q} )
    //
    // to be evaluated at point P, where A = (X-2)/3 and B = (X+1)/3 = A + 1.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN A  = DIVIUEXACT(ADDIS(X, -2), 3);
    int vn = varn(gel(Q, 1));
    
    //
    // Compute f_{A,P}, f_{-A,P}, f_{B,P}, f_{-B,P} and needed powers.
    //
    GEN FA_Q  = pdata->miller_func_f2_f1(pdata, A, Q, P);
    GEN A_Q   = gcopy(pdata->point);
    GEN VA_Q  = pdata->eval_vert_f2_f1(pdata, A_Q, P);
    GEN FMA_Q = FPXQ_INV(FPXQ_MUL(FA_Q, VA_Q, TT, p), TT, p);
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f2_f1(pdata, Q, P);
    
    GEN B_Q   = gcopy(pdata->point);
    GEN FB_Q  = FPXQ_MUL(FA_Q, pdata->num, TT, p); 
        FB_Q  = FPXQ_DIV(FB_Q, pdata->denom, TT, p); 
    GEN VB_Q  = gcopy(pdata->denom);
    GEN FMB_Q = FPXQ_INV(FPXQ_MUL(FB_Q, VB_Q, TT, p), TT, p);
    
    GEN FMA_Q5 = _FpXQ_frobenius_power(pdata, FMA_Q, 5);
    GEN FMA_Q4 = _FpXQ_frobenius_power(pdata, FMA_Q, 4);
    GEN FMB_Q3 = _FpXQ_frobenius_power(pdata, FMB_Q, 3);
    GEN FB_Q2  = _FpXQ_frobenius_power(pdata, FB_Q,  2);
    GEN FB_Q1  = _FpXQ_frobenius_power(pdata, FB_Q,  1);
    
    //
    // Compute needed points
    //
    GEN MA_QP5     = _FpXQ_point_frobenius_power(pdata, A_Q, 5);
    GEN MA_QP4     = _FpXQ_point_frobenius_power(pdata, A_Q, 4);
    gel(MA_QP5, 2) = FPX_NEG(gel(MA_QP5, 2), p);
    gel(MA_QP4, 2) = FPX_NEG(gel(MA_QP4, 2), p);
    
    GEN MB_QP3     = _FpXQ_point_frobenius_power(pdata, B_Q, 3);
    gel(MB_QP3, 2) = FPX_NEG(gel(MB_QP3, 2), p);
    
    GEN B_QP2      = _FpXQ_point_frobenius_power(pdata, B_Q, 2);
    GEN B_QP1      = _FpXQ_point_frobenius_power(pdata, B_Q, 1);
    
    //
    // Compute product of lines divided by product of verticals
    //
    pdata->point = gcopy(MA_QP5);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    
    pdata->update_chord_f2_f1(pdata, MA_QP4, P);
    pdata->update_chord_f2_f1(pdata, MB_QP3, P);
    pdata->update_chord_f2_f1(pdata, B_QP2,  P);
    pdata->update_chord_f2_f1(pdata, B_QP1,  P);
    pdata->update_chord_f2_f1(pdata, A_Q,    P);

    GEN opti = FPXQ_DIV(pdata->num, pdata->denom, TT, p);
        opti = FPXQ_MUL(opti, FA_Q,   TT, p);
        opti = FPXQ_MUL(opti, FMA_Q5, TT, p);
        opti = FPXQ_MUL(opti, FMA_Q4, TT, p);
        opti = FPXQ_MUL(opti, FMB_Q3, TT, p);
        opti = FPXQ_MUL(opti, FB_Q2,  TT, p);
        opti = FPXQ_MUL(opti, FB_Q1,  TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_7_1333(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [1, 0, 0, 0, 0, X]
    //
    // yields the following optimal pairing function:
    //     f_{X,Q}^{q^5} * l_{(X * q^5)*Q, Q} / v_{(1 + X * q^5)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));

    //
    // Computing f_{X, Q}^{p^5}
    //
    GEN QP5 = _FpXQ_point_frobenius_power(pdata, Q, 5);
    GEN FXQP5 = pdata->miller_func_f2_f1(pdata, X, QP5, P);

    //
    // Computing l_{(X * q^5)*Q, Q} / v_{(1 + X * q^5)*Q
    //
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN opti = FPXQ_MUL(FXQP5, pdata->num, TT, p);
    opti = FPXQ_DIV(opti, pdata->denom, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_5_1500(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // NOTE: This is actually the same function as opti_25_1111
    //
    // Using the following short vector:
    //     [X^2, -X, 1, 0]
    //
    // yields the following optimal pairing function:
    //    f_{X^2,Q} * f_{-X,Q}^{q} * l_{(-X*q + q^2)*Q, X^2*Q}
    //    * l_{q^2*Q, -X*q*Q} / (v_{(X^2 + -X*q + q^2)*Q} * v_{(-X* q + q^2)*Q})
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));
    //
    // f_{-X,Q} is computed as 1 / (f{X, Q} * v{XQ})
    //
    GEN FX_Q  = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q   = gcopy(pdata->point);
    GEN FMX_Q = FPXQ_MUL(FX_Q, pdata->eval_vert_f2_f1(pdata, X_Q, P), TT, p);
    FMX_Q = FPXQ_INV(FMX_Q, TT, p);

    GEN MX_Q = gcopy(X_Q);
    gel(MX_Q, 2) = FPX_NEG(gel(X_Q, 2), p);
    
    GEN MX_Q_P  = _FpXQ_point_frobenius_power(pdata, MX_Q, 1);
    GEN FMX_Q_P = _FpXQ_frobenius_power(pdata, FMX_Q, 1);

    //
    // f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN Q_P2   = _FpXQ_point_frobenius_power(pdata,  Q, 2);

    //
    // Computing l_{(p^2)*Q, -X*p*Q} and v_{(p^2 - X*p)*Q}
    //
    pdata->point = gcopy(MX_Q_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q_P2, P);

    //
    // Computing l_{(p^2 - X*p)*Q, X^2*Q}  and v_{(X^2 - X*p + p^2)*Q
    //
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN L = pdata->num;
    GEN V = pdata->denom;

    GEN opti = FPXQ_MUL(FX2_Q, FMX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, L, TT, p);
    opti = FPXQ_DIV(opti, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_8_1250(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // NOTE: This is actually the same function as opti_25_1111
    //
    // Using the following short vector:
    //     [X^2, X, 1, 0]
    //
    // yields the following optimal pairing function:
    //     f_{X^2,Q} * f_{X,Q}^{q} * l_{(X*p + p^2)*Q, X^2*Q}
    //                             * l_{(p^2)*Q, X*p*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q, 1));
    
    //
    // NOTE: f_{X^2,Q} is computed as f{X, Q}^{X} * f{X, XQ}
    //
    GEN FX_Q   = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q    = gcopy(pdata->point);
    GEN FX_XQ  = pdata->miller_func_f2_f1(pdata, X, X_Q, P);
    GEN X2_Q   = gcopy(pdata->point);
    GEN FX_Q_X = FPXQ_POW(FX_Q, X, TT, p);
    GEN FX2_Q  = FPXQ_MUL(FX_Q_X, FX_XQ, TT, p);
    GEN FX_Q_P = _FpXQ_frobenius_power(pdata, FX_Q, 1);
    GEN XP_Q   = _FpXQ_point_frobenius_power(pdata, X_Q, 1);
    GEN P2_Q   = _FpXQ_point_frobenius_power(pdata, Q, 2);

    //
    // Computing l_{(p^2)*Q, X*p*Q}
    //
    pdata->point = XP_Q;
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, P2_Q, P);

    //
    // Computing l_{(p^2 + X*p)*Q, X^2*Q}
    //
    pdata->update_chord_f2_f1(pdata, X2_Q, P);

    GEN opti = FPXQ_MUL(FX2_Q, FX_Q_P, TT, p);
    opti = FPXQ_MUL(opti, pdata->num, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_freeman(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [X, 2*X+1, X, 1]
    //
    // yields the following optimal pairing function:
    //     f_{X,Q} * f_{2*X+1,Q}^p * f_{X,Q}^{p^2}
    //             * l_{(X*p^2 + p^3)*Q, (2*X+1)*p*Q}
    //             * l_{((2*X+1)*p + X*p^2 + p^3)*Q, X*Q}
    //             * l_{(p^3)*Q, X*p^2*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    int vn = varn(gel(Q,1));

    GEN FX_Q    = pdata->miller_func_f2_f1(pdata, X, Q, P);
    GEN X_Q     = gcopy(pdata->point);
    GEN FX_Q_P2 = _FpXQ_frobenius_power(pdata, FX_Q, 2);
    
    //
    // Compute f_{2*X+1,Q} as f_{X,Q}^{2} * l_{X*Q, X*Q} * l{2*X*Q, Q}
    //
    pdata->point = gcopy(X_Q);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f2_f1(pdata, X_Q, P);
    
    GEN F2_X_Q  = gcopy(pdata->num);
    GEN FX_Q_2  = FPXQ_POW(FX_Q, stoi(2), TT, p);
    GEN F2X_Q   = FPXQ_MUL(F2_X_Q, FX_Q_2, TT, p);
    GEN _2X_Q   = gcopy(pdata->point);

    pdata->point = gcopy(_2X_Q);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, Q, P);

    GEN _2XP1_Q   = gcopy(pdata->point);    
    GEN F2XP1_Q   = FPXQ_MUL(F2X_Q, pdata->num, TT, p);    
    GEN F2XP1_Q_P = _FpXQ_frobenius_power(pdata, F2XP1_Q, 1);
    
    //
    //  Computing    l_{(p^3)*Q, (X*p^2)*Q}
    //             * l_{(X*p^2 + p^3)*Q, (2*X+1)*p*Q}
    //             * l_{((2*X+1)*p + X*p^2 + p^3)*Q, X*Q}
    //
    GEN Q_P3      = _FpXQ_point_frobenius_power(pdata,  Q, 3);
    GEN X_Q_P2    = _FpXQ_point_frobenius_power(pdata, X_Q, 2);
    GEN _2XP1_Q_P = _FpXQ_point_frobenius_power(pdata, _2XP1_Q, 1);

    pdata->point = gcopy(Q_P3);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, X_Q_P2, P);
    pdata->update_chord_f2_f1(pdata, _2XP1_Q_P, P);
    pdata->update_chord_f2_f1(pdata, X_Q, P);

    GEN opti = FPXQ_MUL(FX_Q, FX_Q_P2, TT, p);
    opti = FPXQ_MUL(opti, F2XP1_Q_P, TT, p);
    opti = FPXQ_MUL(opti, pdata->num, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_ate_bn(pairing_data_t* pdata, GEN Q, GEN P) {
    //
    // Using the following short vector:
    //     [6*x + 2, 1, −1, 1]
    //
    // yields the following optimal pairing function:
    //    f_{6x + 2,Q} * l_{p^3*Q, −p^2*Q}
    //                 * l_{(p^3 − p^2)*Q, p*Q}
    //                 * l_{(p − p^2 + p^3)*Q, (6x + 2)*Q}
    //
    // to be evaluated at point P.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN Y  = ADDIS(MULUI(6, X), 2);
    int vn = varn(gel(Q,1));

    //
    // Computing [p]*Q, [-p^2]*Q and [p^3]*Q
    //
    GEN QP   = _FpXQ_point_frobenius_power(pdata, Q, 1);
    GEN MQP2 = _FpXQ_point_frobenius_power(pdata, Q, 2);
    GEN QP3  = _FpXQ_point_frobenius_power(pdata, Q, 3);

    gel(MQP2, 2) = FPX_NEG(gel(MQP2, 2), p);

    //
    // Computing f_{6*X+2, Q}
    //
    GEN FYQP = pdata->miller_func_f2_f1(pdata, Y, Q, P);

    //
    // Keeping [6*X+2]*Q
    //
    GEN YQ = gcopy(pdata->point);

    //
    //  Computing   l_{p^3*Q, −p^2*Q}
    //            * l_{(p^3 − p^2)*Q, p*Q}
    //            * l_{(p − p^2 + p^3)*Q, (6x + 2)*Q}
    //
    pdata->point = gcopy(QP3);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);

    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f2_f1(pdata, MQP2, P);
    pdata->update_chord_f2_f1(pdata, QP, P);
    pdata->update_chord_f2_f1(pdata, YQ, P);

    GEN L = pdata->num;

    GEN opti = FPXQ_MUL(FYQP, L, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_duan_et_al_15_1500(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [X^4 - X^3 + X - 1, -X^3 + X^2 - 1]
    //
    // yields the following optimal pairing function:
    //     f_{(-X^4+X^3-X+1), P} * f_{(X^3-X^2+1), P}^{p^5}
    //                           * v{(-X^4+X^3-X+1)*P}
    //
    // to be evaluated at point Q.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X3 = MULII(X2, X);
    GEN Z  = ADDIS(SUBII(X3, X2), 1);
    int vn = varn(gel(Q,1));

    GEN FZ_P    = pdata->miller_func_f1_f2(pdata, Z, P, Q);
    GEN FZ_P_P5 = _FpXQ_frobenius_power(pdata, FZ_P, 5);
    GEN Z_P     = gcopy(pdata->point);

    //
    // f_{X^4-X^3+X, P} computed as f_{X^3-X^2+1, P}^{X} * f_{X, (X^3-X^2+1)*P}
    //
    GEN FZ_P_X = FPXQ_POW(FZ_P, X, TT, p);
    GEN FX_Z_P = pdata->miller_func_f1_f2(pdata, X, Z_P, Q);
    GEN XZ_P   = gcopy(pdata->point);
    GEN FXZ_P  = FPXQ_MUL(FX_Z_P, FZ_P_X, TT, p);

    //
    // f_{-X^4+X^3-X, P} = 1 / (f_{X^4-X^3+X, P} * v_{(X^4-X^3+X)*P})
    //
    GEN VXZ_P  = pdata->eval_vert_f1_f2(pdata, XZ_P, Q);
    GEN FMXZ_P = FPXQ_INV(FPXQ_MUL(FXZ_P, VXZ_P, TT, p), TT, p);

    //
    // f_{-X^4+X^3-X+1, P} computed as
    //     f_{-X^4+X^3-X, P} * l_{(-X^4+X^3-X)*P, P} / v_{(-X^4+X^3-X+1)*P}
    //
    pdata->point = gcopy(XZ_P);
    gel(pdata->point, 2) = FP_NEG(gel(XZ_P, 2), p);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f1_f2(pdata, P, Q);
    
    GEN FMXZP_P = FPXQ_MUL(FMXZ_P, pdata->num, TT, p);
    GEN opti    = FPXQ_MUL(FMXZP_P, FZ_P_P5, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_bn(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [2*X + 1, 6*X^2 + 2*X]
    //
    // yields the following optimal pairing function:
    //     f_{2*X + 1, P} * f_{6*X^2 + 2*X, P}^{p^2}
    //
    // to be evaluated at point Q.
    //
    GEN TT  = pdata->irredpol;
    GEN p   = pdata->charac;
    GEN X   = pdata->u;
    GEN _2X = MULUI(2, X);
    int vn  = varn(gel(Q, 1));

    //
    // f_{6*X^2 + 2*X, P} computed as f_{3*X + 1, P}^{2X} * f_{2X, (3X+1)*P}
    //
    GEN FX_P   = pdata->miller_func_f1_f2(pdata, X, P, Q);
    GEN X_P    = gcopy(pdata->point);
    GEN FX_P_2 = FPXQ_SQR(FX_P, TT, p);
    GEN FX_P_3 = FPXQ_MUL(FX_P, FX_P_2, TT, p);

    GEN F2_X_P = pdata->miller_func_f1_f2(pdata, stoi(2), X_P, Q);
    GEN _2X_P  = gcopy(pdata->point);
    GEN F2X_P  = FPXQ_MUL(F2_X_P, FX_P_2, TT, p);
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f1_f2(pdata, X_P, Q);
    
    GEN F3_X_P = FPXQ_MUL(pdata->num, F2_X_P, TT, p);
    GEN F3X_P  = FPXQ_MUL(F3_X_P, FX_P_3, TT, p);

    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f1_f2(pdata, P, Q);
    
    GEN F3XP_P = FPXQ_MUL(F3X_P, pdata->num, TT, p);
    GEN _3XP_P = gcopy(pdata->point);

    pdata->point = gcopy(_2X_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f1_f2(pdata, P, Q);
    
    GEN F2XP_P    = FPXQ_MUL(F2X_P, pdata->num, TT, p);
    GEN F2X_3XP_P = pdata->miller_func_f1_f2(pdata, _2X, _3XP_P, Q);
    GEN F3XP_P_2X = FPXQ_POW(F3XP_P, _2X, TT, p);
    GEN F6X2_2X_P = FPXQ_MUL(F2X_3XP_P, F3XP_P_2X, TT, p);
    GEN FZ        = _FpXQ_frobenius_power(pdata, F6X2_2X_P, 2);
    
    GEN opti = FPXQ_MUL(F2XP_P, FZ, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_9_1333(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [(X^3 + 1) / 3, (2 - X^3) / 3]
    //
    // yields the following optimal pairing function:
    //     f_{(X^3+1)/3, P} * f_{(2-X^3)/3, P}^{p^3} * v_{[(X^3+1)/3] P}
    //
    // to be evaluated at point Q.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X3 = MULII(X2, X);
    GEN A  = DIVIUEXACT(ADDIS(X3, 1), 3);    // A = (X^3 + 1) / 3
    int vn = varn(gel(Q, 1));

    //
    // Compute f_{A, P}(Q) and v_{A, P}(Q)
    //
    GEN FA_P = pdata->miller_func_f1_f2(pdata, A, P, Q);
    GEN A_P  = gcopy(pdata->point);
    GEN VA_P = pdata->eval_vert_f1_f2(pdata, A_P, Q);
    GEN FVA  = FPXQ_MUL(FA_P, VA_P, TT, p);
    
    //
    // Compute f_{B, P} as f_{1-A,P} = f_{-A, P} * l_{P, [-A]P} / v_{[1-A]P}
    // with f_{-A, P} = 1 / (f_{A, P} * v_{[A]P})
    //
    pdata->point = gcopy(A_P);
    gel(pdata->point, 2) = FP_NEG(gel(pdata->point, 2), p);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f1_f2(pdata, P, Q);
    
    GEN VV      = FPXQ_MUL(FVA, pdata->denom, TT, p);    
    GEN FB_P    = FPXQ_INV(VV, TT, p);  
        FB_P    = FPXQ_MUL(FB_P, pdata->num, TT, p);
    GEN FB_P_Q3 = _FpXQ_frobenius_power(pdata, FB_P, 3);
    
    GEN opti = FPXQ_MUL(FVA, FB_P_Q3, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_18_1333(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [18*(X/7)^3 + 1, -(X/7)^3]
    //
    // yields the following optimal pairing function:
    //     f_{18*(X/7)^3 + 1, P} * f_{-(X/7)^3, P}^{p^6}
    //
    // to be evaluated at point Q.
    //
    GEN TT  = pdata->irredpol;
    GEN p   = pdata->charac;
    GEN X   = pdata->u;
    GEN X7  = DIVIUEXACT(X, 7);
    GEN X7C = MULII(SQRI(X7), X7);
    int vn  = varn(gel(Q, 1));
    
    //
    // Compute f_{18*(X/7)^3+1, P} as f_{18*(X/7)^3, P} * l_{P, (18*(X/7)^3)*P}
    // and f_{18*(X/7)^3, P} as f_{(X/7)^3, P}^{18} * f_{18, ((X/7)^3)*P}
    //
    GEN FX7C_P    = pdata->miller_func_f1_f2(pdata, X7C, P, Q);
    GEN X7C_P     = gcopy(pdata->point);
    GEN FMX7C_P   = FPXQ_INV(FX7C_P, TT, p);
    GEN FX7C_P_18 = FPXQ_POW(FX7C_P, stoi(18), TT, p);
    GEN F18_X7C_P = pdata->miller_func_f1_f2(pdata, stoi(18), X7C_P, Q);
    GEN F18X7C_P  = FPXQ_MUL(F18_X7C_P, FX7C_P_18, TT, p);
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);

    pdata->update_chord_f1_f2(pdata, P, Q);
    
    GEN F18X7CP_P  = FPXQ_MUL(pdata->num, F18X7C_P, TT, p);
    GEN FMX7C_P_P5 = _FpXQ_frobenius_power(pdata, FMX7C_P, 6);
    
    GEN opti = FPXQ_MUL(F18X7CP_P, FMX7C_P_P5, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_16_1250(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // WARNING: I am not sure if this vector is always correct!
    //
    // Using the following short vector:
    //     [49 * X^4 / 625, 7 + 168 * X^4 / 625]
    //
    // yields the following optimal pairing function:
    //      f_{(49*X^4/625),P} * f_{(7+168*X^4/625),P}^{p^4}
    //
    // to be evaluated at point Q.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X4 = SQRI(X2);
    GEN Y  = DIVIUEXACT(X4, 625);
    int vn  = varn(gel(Q, 1));
    
    //
    // f_{49*X^4/625,P} computed as f_{X^4/625,P}^49 * f_{49,[X^4/625]P}
    //
    GEN FY   = pdata->miller_func_f1_f2(pdata, Y, P, Q);
    GEN YP   = gcopy(pdata->point);
    GEN FY49 = FPXQ_POW(FY, stoi(49), TT, p);
    GEN FA   = pdata->miller_func_f1_f2(pdata, stoi(49), YP, Q);
    FA = FPXQ_MUL(FA, FY49, TT, p);
    
    //
    // f_{168*X^4/625,P} computed as f_{X^4/625,P}^168 * f_{168,[X^4/625]P}
    //
    GEN E2 = FPXQ_POW(FY, stoi(168), TT, p);
    GEN FB = pdata->miller_func_f1_f2(pdata, stoi(168), YP, Q);
    GEN ZP = gcopy(pdata->point);
    FB = FPXQ_MUL(FB, E2, TT, p);
    
    //
    // f_{7+168*X^4/625,P} computed as
    //      f_{168*X^4/625,P} * f_{7,P} * l_{[7]P, [168*X^4/625]P}
    //
    GEN FS = pdata->miller_func_f1_f2(pdata, stoi(7), P, Q);
    
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f1_f2(pdata, ZP, Q);
    
    FB = FPXQ_MUL(FB, pdata->num, TT, p);
    FB = FPXQ_MUL(FB, FS, TT, p);
    
    GEN FBP4 = _FpXQ_frobenius_power(pdata, FB, 4);
    GEN opti = FPXQ_MUL(FA, FBP4, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_15_1500(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [X^3 + X^2 - 1, X^4 + X^3 - X - 1]
    //
    // yields the following optimal pairing function:
    //     f_{(X^3 + X^2 - 1), P} * f_{(X^4 + X^3 -X - 1), P}^{p^5}
    //                            * v{(X^3 + X^2 - 1)*P}
    //
    // to be evaluated at point Q.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X3 = MULII(X2, X);
    GEN Y  = ADDIS(ADDII(X3, X2), -1);
    int vn  = varn(gel(Q, 1));
        
    GEN FY_P   = pdata->miller_func_f1_f2(pdata, Y, P, Q);
    GEN Y_P    = gcopy(pdata->point);
    GEN V1     = pdata->eval_vert_f1_f2(pdata, Y_P, Q);
    GEN FY_P_X = FPXQ_POW(FY_P, X, TT, p);
    GEN FX_Y_P = pdata->miller_func_f1_f2(pdata, X, Y_P, Q);
    GEN XY_P   = gcopy(pdata->point);
    GEN FXY_P  = FPXQ_MUL(FX_Y_P, FY_P_X, TT, p);
    
    GEN MP = gcopy(P);
    gel(MP, 2) = FP_NEG(gel(P, 2), p);
    GEN V2 = pdata->eval_vert_f1_f2(pdata, MP, Q);
    
    pdata->point = gcopy(XY_P);
    pdata->num   = mkpoln(1, gen_1);
    pdata->denom = mkpoln(1, gen_1);
    setvarn(pdata->num, vn);
    setvarn(pdata->denom, vn);
    
    pdata->update_chord_f1_f2(pdata, MP, Q);
    
    V2 = FPXQ_MUL(V2, pdata->denom, TT, p);
    
    GEN FXYM_P = FPXQ_MUL(FXY_P, pdata->num, TT, p);
    FXYM_P = FPXQ_DIV(FXYM_P, V2, TT, p);
    
    GEN FXYM_P_P5 = _FpXQ_frobenius_power(pdata, FXYM_P, 5);
    
    GEN opti = FPXQ_MUL(FY_P, FXYM_P_P5, TT, p);
    opti = FPXQ_MUL(opti, V1, TT, p);
    
    return opti;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_24_1250(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [ X^4, -1]
    //
    // yields the following optimal pairing function:
    //     f_{X^4, P}
    //
    // to be evaluated at point Q.
    //
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X4 = SQRI(X2);

    GEN FX4P = pdata->miller_func_f1_f2(pdata, X4, P, Q);

    return FX4P;
}
//----------------------------------------------------------------------------
GEN
opti_twisted_27_1111(pairing_data_t* pdata, GEN P, GEN Q) {
    //
    // Using the following short vector:
    //     [X^9, 1]
    //
    // yields the following optimal pairing function:
    //     f_{X^9, P} * v{(X^9)*P}
    //
    // to be evaluated at point Q.
    //
    GEN TT = pdata->irredpol;
    GEN p  = pdata->charac;
    GEN X  = pdata->u;
    GEN X2 = SQRI(X);
    GEN X4 = SQRI(X2);
    GEN X8 = SQRI(X4);
    GEN X9 = MULII(X8, X);

    GEN FX9P = pdata->miller_func_f1_f2(pdata, X9, P, Q);
    GEN V    = pdata->eval_vert_f1_f2(pdata, pdata->point, Q);
    GEN opti = FPXQ_MUL(FX9P, V, TT, p);

    return opti;
}
//----------------------------------------------------------------------------
