#include <stdint.h>
#include "cyclo_poly.h"
#include "funcs.h"

//
// Returns 0 if the I-th bit of the ulong N is set to zero and 2^I otherwise.
// I should be in [0, 31].
//
#define GET_ITH_BIT(N, I) ( (N) & ((uint64_t)1 << (I)) )
#define NCYCLOTOMIC_POLY 63
//----------------------------------------------------------------------------
struct cyclo_poly_struct {
    uint64_t posi;
    uint64_t nega;
};
typedef struct cyclo_poly_struct cyclo_poly_t;
//----------------------------------------------------------------------------
cyclo_poly_t
cyclotomic_polys[] = {
    {0, 0},
    {2, 1},
    {3, 0},
    {7, 0},
    {5, 0},
    {31, 0},
    {5, 2},
    {127, 0},
    {17, 0},
    {73, 0},
    {21, 10},
    {2047, 0},
    {17, 4},
    {8191, 0},
    {85, 42},
    {297, 146},
    {257, 0},
    {131071, 0},
    {65, 8},
    {524287, 0},
    {273, 68},
    {4681, 2322},
    {1365, 682},
    {8388607, 0},
    {257, 16},
    {1082401, 0},
    {5461, 2730},
    {262657, 0},
    {4369, 1092},
    {536870911, 0},
    {387, 56},
    {2147483647, 0},
    {65537, 0},
    {1198665, 599186},
    {87381, 43690},
    {17454241, 8726850},
    {4097, 64},
    {137438953471, 0},
    {349525, 174762},
    {19173961, 9585810},
    {65793, 4112},
    {2199023255551, 0},
    {6211, 792},
    {8796093022207, 0},
    {1118481, 279620},
    {16810497, 2101256},
    {5592405, 2796202},
    {140737488355327, 0},
    {65537, 256},
    {4432676798593, 0},
    {1049601, 32800},
    {4908552777, 2454267026},
    {17895697, 4473924},
    {9007199254740991, 0},
    {262145, 512},
    {1134979744801, 567212642370},
    {16843009, 1052688},
    {78536544841, 39268197522},
    {357913941, 178956970},
    {576460752303423487, 0},
    {81925, 1344},
    {2305843009213693951, 0},
    {1431655765, 715827882},
    {68853957121, 8606715912}
};
//----------------------------------------------------------------------------
GEN
get_ith_cyclotomic_poly(unsigned int i)
{
    if (i >= NCYCLOTOMIC_POLY) {
        pariprintf("cannot evaluate %u-th cyclotomic polynomial", i);
        pariflush();
        return gen_0;
    }
    uint64_t posi = cyclotomic_polys[i].posi;
    uint64_t nega = cyclotomic_polys[i].nega;
    
    GEN poly  = cgetg(i + 2, t_POL);
    GEN coefs = poly + 2;
    
    for (unsigned int j = 0; j <= i; j++) {
        if (GET_ITH_BIT(posi, j) != 0) {
            gel(coefs, j) = stoi(1);
        
        } else {
            if (GET_ITH_BIT(nega, j) != 0) {
                gel(coefs, j) = stoi(-1);
            } else {
                gel(coefs, j) = stoi(0);
            }
        }
    }
    poly[1] = evalvarn(0);
    return normalizepol_lg(poly, i + 2);
}
//----------------------------------------------------------------------------
GEN
eval_ith_cyclotomic_poly(unsigned int i, GEN x)
{
    IN();
    if (i >= NCYCLOTOMIC_POLY) {
        pariprintf("cannot evaluate %u-th cyclotomic polynomial", i);
        pariflush();
        return NULL;
    }    
    uint64_t posi = cyclotomic_polys[i].posi;
    uint64_t nega = cyclotomic_polys[i].nega;
    
    GEN xpow = stoi(1);
    GEN res  = stoi(0);
    
    for (unsigned int j = 0; j <= i; j++) {
        if (GET_ITH_BIT(posi, j) != 0) {
            res = addii(res, xpow);
        }
        if (GET_ITH_BIT(nega, j) != 0) {
            res = subii(res, xpow);
        }
        xpow = mulii(xpow, x);
    }
    OUT();
    return res;
}
//----------------------------------------------------------------------------
int
get_ith_cyclotomic_poly_degree(unsigned int i)
{
    IN();
    if (i >= NCYCLOTOMIC_POLY) {
        pariprintf("cannot evaluate %u-th cyclotomic polynomial", i);
        pariflush();
        return 0;
    }
    int deg = 0;
    uint64_t posi = cyclotomic_polys[i].posi;
    for (int j = i; j >= 0; j--) {
        if (GET_ITH_BIT(posi, j) != 0) {
            deg = j;
            break;
        }
    }
    OUT();
    return deg;
}
//----------------------------------------------------------------------------
