#if !defined(_APIP_PAIRINGS_H_)
#define _APIP_PAIRINGS_H_

#include "apip_config.h"

#ifdef __cplusplus
extern "C" {
#endif

//----------------------------------------------------------------------------
#if defined(__GNUC__)
    #ifndef MAYBE_UNUSED
        #define MAYBE_UNUSED __attribute__ ((unused))
    #endif
    #ifndef IS_LIKELY
        #define IS_LIKELY(X) __builtin_expect(X, 1)
    #endif
    #ifndef IS_UNLIKELY
        #define IS_UNLIKELY(X) __builtin_expect(X, 0)
    #endif
#else
    #ifndef MAYBE_UNUSED
        #define MAYBE_UNUSED /* intentionally left empty */
    #endif
    #ifndef IS_LIKELY
        #define IS_LIKELY(X) (X)
    #endif
    #ifndef IS_UNLIKELY
        #define IS_UNLIKELY(X) (X)
    #endif
#endif
//----------------------------------------------------------------------------
#define ERROR   0
#define SUCCESS 1
#define _DEBUG_ 0 
                //APIP_DEBUG
#if _DEBUG_ != 0
    //
    // A few debugging macros:
    //   MSG -- Prints a debug message
    //    IN -- Prints a correctly indented message when entering a function
    //   OUT -- Prints a correctly indented message when exiting a function
    //
    MAYBE_UNUSED static int _indent = 0;
    #define MSG(M);                                                     \
        pariprintf("%s: %s\n", __func__, M);                            \
        pariflush();
    #define IN()                                                        \
        do {                                                            \
            int i;                                                      \
            _indent++;                                                  \
            for (i = 0; i < _indent; i++) {pariprintf("  ");}           \
            pariprintf("%s: entered\n", __func__);                      \
            pariflush();                                                \
        } while (0)
    #define OUT()                                                       \
        do {                                                            \
            int i;                                                      \
            for (i = 0; i < _indent; i++) {pariprintf("  ");}           \
            pariprintf("%s: exited\n", __func__);                       \
            pariflush();                                                \
            _indent--;                                                  \
        } while (0)
#else
    #define MSG(M)      /* intentionally left blank */
    #define IN()        /* intentionally left blank */
    #define OUT()       /* intentionally left blank */
#endif
#define IS_EVEN(ui)          ( ((ui) & 1) == 0 )
#define IS_ODD(ui)           ( ((ui) & 1) != 0 )
#define IS_POWER_OF_2_UI(ui) ( ((ui) & ((ui) - 1)) == 0 )
#define ERR(X)          pariprintf("%s:%s: " X "\n", __FILE__, __func__);
#define NTINY_PRIMES    20
//----------------------------------------------------------------------------
static short tiny_primes[NTINY_PRIMES] MAYBE_UNUSED = {
    2,   3,   5,   7,  11,  13,  17,  19,  23,  29, 
   31,  37,  41,  43,  47,  53,  59,  61,  67,  71
};
//----------------------------------------------------------------------------
typedef struct {
    //
    // A structure used for computation of pairings. It combines a point on
    // an elliptic curve and the evaluation of some function given by its
    // numerator num and denominator denom...
    //
    GEN point;
    GEN num;
    GEN denom;
    //
    // ... together with some useful (cached) data.
    //
    GEN charac;
    GEN defpol;
    int edegree;
} paf;
//----------------------------------------------------------------------------
GEN   _ZX_muls(GEN y, long x);
GEN   _FpXQ_to_FF(GEN v, GEN T, GEN p);
GEN   _Flxq_to_FF(GEN v, GEN T, long l);
GEN   _F2xq_to_FF(GEN v, GEN T);
GEN   _intmod_val(GEN x);
ulong _get_bit(GEN x, long n);
GEN   _get_char(GEN a);
int*  _naf(GEN k);
//----------------------------------------------------------------------------
int   _are_same_point(GEN A, GEN B);
int   _are_opposite_points(GEN E, GEN A, GEN B);
int   _ZX_equal(GEN x, GEN y);
//----------------------------------------------------------------------------

#ifdef __cplusplus
}
#endif
#endif
