#if !defined(_APIP_PAIRING_DATA_H_)
#define _APIP_PAIRING_DATA_H_

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#if !defined(_PARIPRIV_H_)
    #define _PARIPRIV_H_
    #include <paripriv.h>
#endif

#include <strings.h>

#define N_CHARAC_CODES             3
#define N_CURVE_FAMILIES          23
#define N_COORD_SYSTEMS            3
#define N_MILLER_METHODS           5
#define N_HARDPOW_METHODS          5
#define MAX_EMBED_DEGREE          63
#define MIN_NFROB_POWERS           6

#define MAX_NFROB_POWERS          (MAX_EMBED_DEGREE / 2)
#define NAIVE_HARD_POW_THRESHOLD  MAX_NFROB_POWERS

#define DFLT_FROB_COST  0.75
#define DFLT_SQR_COST   0.90

#define PRINT_CHARAC_CODE(CODE)    printf("%s\n", charac_codes_to_str[CODE]);
#define PRINT_CURVE_FAMILY(CODE)   printf("%s\n", curve_families_to_str[CODE]);
#define PRINT_COORD_SYSTEM(CODE)   printf("%s\n", coord_systems_to_str[CODE]);
#define PRINT_MILLER_METHOD(CODE)  printf("%s\n", miller_methods_to_str[CODE]);
#define PRINT_HARDPOW_METHOD(CODE) printf("%s\n", hardpow_methods_to_str[CODE]);

enum charac_code_enum {
    CHARAC_2,
    CHARAC_3,
    CHARAC_P
};
typedef enum charac_code_enum charac_code_t;

enum curve_family_enum {
    SS2,
    SS3,
    MNT,
    BN,
    SCOTT_BARRETO_6,
    FREEMAN,
    DUAN_ET_AL_15_1500,
    OPTI_5_1500,
    OPTI_7_1333,
    OPTI_8_1250,
    OPTI_9_1333,
    OPTI_11_1200,
    OPTI_15_1500,
    OPTI_16_1250,
    OPTI_17_1125,
    OPTI_17_1188,
    OPTI_18_1333,
    OPTI_19_1111,
    OPTI_24_1250,
    OPTI_25_1300,
    OPTI_26_1166,
    OPTI_27_1111,
    UNKNOWN
};
typedef enum curve_family_enum curve_family_t;

enum coord_system_enum {
    AFFINE,
    PROJECTIVE,
    JACOBIAN
};
typedef enum coord_system_enum coord_system_t;

enum miller_method_enum {
    STANDARD,
    NAF,
    SLIDING_WINDOW,
    BOXALL_ET_AL,
    BOXALL_ET_AL_NAF
};
typedef enum miller_method_enum miller_method_t;

enum hardpow_method_enum {
    NAIVE,
    INTERLEAVING,
    AVANZI_MIHAILESCU,
    KATO_ET_AL,
    SCOTT_ET_AL
};
typedef enum hardpow_method_enum hardpow_method_t;

struct pairing_data_struct {
    GEN curve;
    GEN charac;
    int charac_code;
    int f1_extdeg;
    int f2_extdeg;
    GEN g1_order;
    GEN frob_trace;

    curve_family_t   family;
    coord_system_t   coordsys;
    miller_method_t  miller_method;
    hardpow_method_t hardpow_method;

    unsigned short embed_deg;
    unsigned short edeg_sdiv;

    unsigned short card_aut_E;
    unsigned short twist_deg;
    unsigned short m;
    unsigned short e;

    GEN final_exp;
    GEN easy_exp_2;
    GEN easy_exp_1;
    GEN hard_exp;

    unsigned short do_denom_elim;
    unsigned short fields_are_friendly;
    unsigned short char_is_1_mod_12;
    unsigned short embed_deg_is_2i3j;
    unsigned short precomp_done;
    unsigned short do_reduce;
    unsigned short fmaps_are_computed;
    unsigned short do_naive_exp;
    unsigned short do_naive_hard_exp;

    GEN irredpol;
    GEN point;
    GEN num;
    GEN denom;

    GEN  (*miller_func_f1_f2)   (struct pairing_data_struct*, GEN, GEN, GEN);
    GEN  (*miller_func_f2_f1)   (struct pairing_data_struct*, GEN, GEN, GEN);
    GEN  (*opti_ate_func)       (struct pairing_data_struct*, GEN, GEN);
    GEN  (*opti_twisted_func)   (struct pairing_data_struct*, GEN, GEN);
    GEN  (*final_exp_func)      (struct pairing_data_struct*, GEN);

    GEN (*point_mul_func_f1)    (struct pairing_data_struct*, GEN, GEN);
    GEN (*point_add_func_f1)    (struct pairing_data_struct*, GEN, GEN);
    GEN (*point_double_func_f1) (struct pairing_data_struct*, GEN);

    GEN (*point_mul_func_f2)    (struct pairing_data_struct*, GEN, GEN);
    GEN (*point_add_func_f2)    (struct pairing_data_struct*, GEN, GEN);
    GEN (*point_double_func_f2) (struct pairing_data_struct*, GEN);

    GEN (*eval_vert_f1_f2)      (struct pairing_data_struct*, GEN, GEN);
    GEN (*eval_vert_f2_f1)      (struct pairing_data_struct*, GEN, GEN);

    void (*update_chord_f1_f2)   (struct pairing_data_struct*, GEN, GEN);
    void (*update_chord_f2_f1)   (struct pairing_data_struct*, GEN, GEN);
    void (*update_tangent_f1_f2) (struct pairing_data_struct*, GEN);
    void (*update_tangent_f2_f1) (struct pairing_data_struct*, GEN);

    GEN to_iso;
    GEN to_iso_inv;
    GEN frobenius_powers[MAX_NFROB_POWERS + 1];

    GEN u;
    GEN cof;
    GEN fe_coefs;

    unsigned short r;
    
    float frobcost;
    float sqrcost;
    
    GEN tmp;
    GEN tmp2;
};
typedef struct pairing_data_struct pairing_data_t;

static const char* const charac_codes_to_str[N_CHARAC_CODES] = {
    "characteristic 2",
    "characteristic 3",
    "characteristic p > 3"
};

static const char* const curve_families_to_str[N_CURVE_FAMILIES] = {
    "supersingular char 2",
    "supersingular char 3",
    "Miyaji-Nakabayashi-Takano",
    "Barreto-Naehrig",
    "Scott-Barreto -- k=6 & rho=1.250",
    "Freeman",
    "Brezing-Weng (Duan et al) -- k=15 & rho=1.500",
    "Cyclotomic curve -- k= 5 & rho=1.500",
    "Cyclotomic curve -- k= 7 & rho=1.333",
    "Cyclotomic curve -- k= 8 & rho=1.250",
    "Cyclotomic curve -- k= 9 & rho=1.333",
    "Cyclotomic curve -- k=11 & rho=1.200",
    "Cyclotomic curve -- k=15 & rho=1.500",
    "Cyclotomic curve -- k=16 & rho=1.250",
    "Cyclotomic curve -- k=17 & rho=1.125",
    "Cyclotomic curve -- k=17 & rho=1.188",
    "Cyclotomic curve -- k=18 & rho=1.333",
    "Cyclotomic curve -- k=19 & rho=1.111",
    "Cyclotomic curve -- k=24 & rho=1.1250",
    "Cyclotomic curve -- k=25 & rho=1.300",
    "Cyclotomic curve -- k=26 & rho=1.166",
    "Cyclotomic curve -- k=27 & rho=1.111",
    "unknown"
};

static const char* const coord_systems_to_str[N_COORD_SYSTEMS] = {
    "affine",
    "projective",
    "jacobian"
};

static const char* const miller_methods_to_str[N_MILLER_METHODS] = {
    "standard",
    "non-adjagent form",
    "sliding window",
    "Boxall et al."
    "Boxall et al. with NAF"
};

static const char* const hardpow_methods_to_str[N_HARDPOW_METHODS] = {
    "naive",
    "interleaving",
    "Avanzi-Mihailescu -- Frobenius abusing algorithm",
    "Kato et al."
    "Scott et al."
};

static inline
coord_system_t charac_codes_from_str(char* code)
{
    if (!strcasecmp(code, "CHARAC_2")) {
        return CHARAC_2;
    }
    if (!strcasecmp(code, "CHARAC_3")) {
        return CHARAC_3;
    }
    if (!strcasecmp(code, "CHARAC_P")) {
        return CHARAC_P;
    }
    return CHARAC_2;
}

static inline
coord_system_t coord_system_from_str(char* system)
{
    if (!strcasecmp(system, "AFFINE")) {
        return AFFINE;
    }
    if (!strcasecmp(system, "PROJECTIVE")) {
        return PROJECTIVE;
    }
    if (!strcasecmp(system, "JACOBIAN")) {
        return JACOBIAN;
    }
    return AFFINE;
}

static inline
miller_method_t miller_method_from_str(char* method)
{
    if (!strcasecmp(method, "STD")) {
        return STANDARD;
    }
    if (!strcasecmp(method, "NAF")) {
        return NAF;
    }
    if (!strcasecmp(method, "SLIDING_WINDOW")) {
        return SLIDING_WINDOW;
    }
    if (!strcasecmp(method, "BOXALL_ET_AL")) {
        return BOXALL_ET_AL;
    }
    if (!strcasecmp(method, "BOXALL_ET_AL_NAF")) {
        return BOXALL_ET_AL_NAF;
    }
    return STANDARD;
}

static inline
hardpow_method_t hardpow_method_from_str(char* method)
{
    if (!strcasecmp(method, "NAIVE")) {
        return NAIVE;
    }
    if (!strcasecmp(method, "INTERLEAVING")) {
        return INTERLEAVING;
    }
    if (!strcasecmp(method, "AVANZI_MIHAILESCU")) {
        return AVANZI_MIHAILESCU;
    }
    if (!strcasecmp(method, "KATO_ET_AL")) {
        return KATO_ET_AL;
    }
    if (!strcasecmp(method, "SCOTT_ET_AL")) {
        return SCOTT_ET_AL;
    }
    return NAIVE;
}

static inline
curve_family_t curve_family_from_str(char* family)
{
    if (!strcasecmp(family, "SS2")) {
        return SS2;
    }
    if (!strcasecmp(family, "SS3")) {
        return SS3;
    }
    if (!strcasecmp(family, "MNT")) {
        return MNT;
    }
    if (!strcasecmp(family, "BN")) {
        return BN;
    }
    if (!strcasecmp(family, "SCOTT_BARRETO_6")) {
        return SCOTT_BARRETO_6;
    }
    if (!strcasecmp(family, "FREEMAN")) {
        return FREEMAN;
    }
    if (!strcasecmp(family, "DUAN_ET_AL_15_1500")) {
        return DUAN_ET_AL_15_1500;
    }
    if (!strcasecmp(family, "OPTI_5_1500")) {
        return OPTI_5_1500;
    }
    if (!strcasecmp(family, "OPTI_7_1333")) {
        return OPTI_7_1333;
    }
    if (!strcasecmp(family, "OPTI_8_1250")) {
        return OPTI_8_1250;
    }
    if (!strcasecmp(family, "OPTI_9_1333")) {
        return OPTI_9_1333;
    }
    if (!strcasecmp(family, "OPTI_11_1200")) {
        return OPTI_11_1200;
    }
    if (!strcasecmp(family, "OPTI_15_1500")) {
        return OPTI_15_1500;
    }
    if (!strcasecmp(family, "OPTI_16_1250")) {
        return OPTI_16_1250;
    }
    if (!strcasecmp(family, "OPTI_17_1125")) {
        return OPTI_17_1125;
    }
    if (!strcasecmp(family, "OPTI_17_1188")) {
        return OPTI_17_1188;
    }
    if (!strcasecmp(family, "OPTI_18_1333")) {
        return OPTI_18_1333;
    }
    if (!strcasecmp(family, "OPTI_19_1111")) {
        return OPTI_19_1111;
    }
    if (!strcasecmp(family, "OPTI_24_1250")) {
        return OPTI_24_1250;
    }
    if (!strcasecmp(family, "OPTI_25_1300")) {
        return OPTI_25_1300;
    }
    if (!strcasecmp(family, "OPTI_26_1166")) {
        return OPTI_26_1166;
    }
    if (!strcasecmp(family, "OPTI_27_1111")) {
        return OPTI_27_1111;
    }
    if (!strcasecmp(family, "UNKNOWN")) {
        return UNKNOWN;
    }
    return UNKNOWN;
}
//----------------------------------------------------------------------------
pairing_data_t*
alloc_pairing_data(
    GEN curve, GEN charac, GEN f1_defpol, GEN f2_defpol,
    GEN g1_order, curve_family_t family
);
void free_pairing_data(pairing_data_t* pdata);
void update_pairing_data(pairing_data_t* pdata);
void set_twist_degree(pairing_data_t* pdata, unsigned int d);
GEN  frobenius_power(pairing_data_t* const pdata, GEN X, unsigned int i);
void compute_frobenius_powers(pairing_data_t* const pdata);
void print_frobenius_matrix(pairing_data_t* const pdata, unsigned int i);
//----------------------------------------------------------------------------
GEN weil(pairing_data_t*, GEN, GEN);
GEN tate(pairing_data_t*, GEN, GEN);
GEN  ate(pairing_data_t*, GEN, GEN);
GEN opti_ate(pairing_data_t*, GEN, GEN);
GEN twisted(pairing_data_t*, GEN, GEN);
GEN opti_twisted(pairing_data_t*, GEN, GEN);
GEN miller(pairing_data_t*, GEN, GEN, GEN);
//----------------------------------------------------------------------------

#ifdef __cplusplus
}
#endif
#endif
