//
// WARNING:
// This file was taken from the TIFA library
//     http://www.lix.polytechnique.fr/Labo/Jerome.Milan/tifa/tifa.xhtml
// and is subject to the LGPL version 2.1 or later.
//

#if !defined(_APIP_TIMER_H_)
#define _APIP_TIMER_H_

#ifdef __cplusplus
extern "C" {
#endif

#if !defined(__TIMING__)
#error "__TIMING__ must be defined before including this header."
#endif

#if __TIMING__
    #include "stopwatch.h"
    
    #define TIMING_FORMAT "%8.4f"
    #define INIT_NAMED_TIMER(NAME)                  \
        stopwatch_t __APIP_STOPWATCH_ ##NAME;       \
        init_stopwatch(&__APIP_STOPWATCH_ ##NAME);

    #define RESET_NAMED_TIMER(NAME)                     \
        do {                                            \
            reset_stopwatch(&__APIP_STOPWATCH_ ##NAME); \
        } while (0)

    #define START_NAMED_TIMER(NAME)                     \
        do {                                            \
            start_stopwatch(&__APIP_STOPWATCH_ ##NAME); \
        } while (0)

    #define STOP_NAMED_TIMER(NAME)                      \
        do {                                            \
            stop_stopwatch(&__APIP_STOPWATCH_ ##NAME);  \
        } while (0)

    #define GET_NAMED_TIMING(NAME)                          \
        get_stopwatch_elapsed(&__APIP_STOPWATCH_ ##NAME)
        
    #define INIT_TIMER       INIT_NAMED_TIMER()
    #define RESET_TIMER      RESET_NAMED_TIMER()
    #define START_TIMER      START_NAMED_TIMER()
    #define STOP_TIMER       STOP_NAMED_TIMER()
    #define GET_TIMING       GET_NAMED_TIMING()
    #define PRINT_TIMING(S)  printf("%s : "TIMING_FORMAT" s\n", S, GET_TIMING);  
#else
    #define INIT_NAMED_TIMER(NAME)    /* intentionally left empty */
    #define RESET_NAMED_TIMER(NAME)   /* intentionally left empty */
    #define START_NAMED_TIMER(NAME)   /* intentionally left empty */
    #define STOP_NAMED_TIMER(NAME)    /* intentionally left empty */
    #define GET_NAMED_TIMING(NAME)    (-1.0)

    #define INIT_TIMER       /* intentionally left empty */
    #define RESET_TIMER      /* intentionally left empty */
    #define START_TIMER      /* intentionally left empty */
    #define STOP_TIMER       /* intentionally left empty */
    #define GET_TIMING       (-1.0)
    #define PRINT_TIMING(S)  /* intentionally left empty */
#endif

#ifdef __cplusplus
}
#endif

#endif
